'''
Created on Dec 2, 2015

@author: xiao
'''
import bisect
import logging
import numpy as np
import pandas as pd
from collections import defaultdict
from functools import cmp_to_key
from pdfminer.utils import Plane
from pdftotree.pdf.vector_utils import inside, reading_order
from pprint import pformat

class Cell(object):
    '''Represents a cell with no visual dividers inside'''
    def __init__(self, origin, texts=[], rowspan=1, colspan=1):
        '''
        origin: the top left grid coordinate of the cell
        '''
        self.rowstart, self.colstart = origin
        self.rowend = self.rowstart + rowspan
        self.colend = self.colstart + colspan
        self.texts = texts

    def __str__(self, *args, **kwargs):
        return ','.join([m.get_text().encode('utf8') for m in self.texts])

class Grid(object):
    '''
    A rendered grid to capture structural layout info
    '''
    def __init__(self, mentions, lines, region, min_cell_size=6.0):
        '''
        Constructor
        '''
        self.min_cell_size = min_cell_size
        vlines, hlines = _split_vlines_hlines(lines)

        self.xs = [v.xc for v in vlines]
        self.ys = [h.yc for h in hlines]

        # Remove closely clustered lines
        # Also make sure there is at least 1 mega column for the table
        self.xs = _retain_centroids(self.xs + [region.x0, region.x1], min_cell_size)
        self.ys = _retain_centroids(self.ys + [region.y0, region.y1], min_cell_size)

        self.xranges = zip(self.xs, self.xs[1:])
        self.yranges = zip(self.ys, self.ys[1:])

        self.num_cols = len(self.xranges)
        self.num_rows = len(self.yranges)

        # Grid contents
        self._grid = np.full([self.num_rows, self.num_cols], None, dtype=np.dtype(object))
        grid = self._grid

        # Record whether a particular cell boundary is present
        line_plane = Plane(region.bbox)
        line_plane.extend(lines)
        vbars, hbars = self._mark_grid_bounds(line_plane, region)
        cells = []
        # Establish cell regions
        for i in xrange(self.num_rows):
            for j in xrange(self.num_cols):
                if grid[i, j]: continue  # Skip already marked cells
                # Merge with cell above
                if i > 0 and not hbars[i, j]:
                    grid[i, j] = cell = grid[i - 1, j]
                    cell.rowend = i + 1
                # Merge with cell left
                elif j > 0 and not vbars[i, j]:
                    grid[i, j] = cell = grid[i, j - 1]
                    cell.colend = j + 1
                # Create new cell otherwise
                else:
                    grid[i, j] = cell = Cell([i, j])
                    cells.append(cell)

        # Now get the cell's contents by using its boundary
        text_plane = Plane(region.bbox)
        text_plane.extend(mentions)

        for cell in cells:
            x0 = self.xs[cell.colstart]
            x1 = self.xs[cell.colend]
            y0 = self.ys[cell.rowstart]
            y1 = self.ys[cell.rowend]
            bbox = (x0, y0, x1, y1)
            # Keep mentions whose centers are inside the cell
            cell.texts = [m for m in text_plane.find(bbox) if inside(bbox, (m.xc, m.yc) * 2)]

        # TODO: provide HTML conversion here

        self.get_normalized_grid()

    def to_dataframe(self):
        return pd.DataFrame(self._grid)

    def to_html(self):
        return self.to_dataframe().to_html(index=False, header=False)

    def get_normalized_grid(self):
        '''
        Analyzes subcell structure
        '''
        log = logging.getLogger(__name__)
        # Resolve multirow mentions, TODO: validate against all PDFs
        subcol_count = 0;
        mega_rows = []
        for row_id, row in enumerate(self._grid):
            # maps yc_grid -> [mentions]
            subrow_across_cell = defaultdict(list)
            for col_id, cell in enumerate(row):
                # Keep cell text in reading order
                cell.texts.sort(key=cmp_to_key(reading_order))
#                intervals, groups = project_onto(cell.texts, axis='x', self.min_cell_size)
                prev = None
                log.debug('='*50)
                for m in cell.texts:
                    subrow_across_cell[m.yc_grid].append(m)
                    prev = m

            log.debug(pformat(dict(subrow_across_cell)))

            mega_rows.append(subrow_across_cell)

        # Multiline paragraph check
        # Subrow/Subcolumn

        return mega_rows

    def _mark_grid_bounds(self, plane, region_bbox):
        '''
        Assume all lines define a complete grid over the region_bbox.
        Detect which lines are missing so that we can recover merged
        cells.
        '''
        # Grid boundaries
        vbars = np.zeros([self.num_rows, self.num_cols + 1], dtype=np.bool)
        hbars = np.zeros([self.num_rows + 1, self.num_cols], dtype=np.bool)

        def closest_idx(arr, elem):
            left = bisect.bisect_left(arr, elem) - 1
            right = bisect.bisect_right(arr, elem) - 1
            return left if abs(arr[left] - elem) < abs(arr[right] - elem) else right

        # Figure out which separating segments are missing, i.e. merge cells
        for row, (y0, y1) in enumerate(self.yranges):
            yc = (y0 + y1) / 2
            for l in plane.find((region_bbox.x0, yc, region_bbox.x1, yc)):
                vbars[row, closest_idx(self.xs, l.xc)] = True
        for col, (x0, x1) in enumerate(self.xranges):
            xc = (x0 + x1) / 2
            for l in plane.find((xc, region_bbox.y0, xc, region_bbox.y1)):
                hbars[closest_idx(self.ys, l.yc), col] = True
        return vbars, hbars

############################
# Utilities
############################

def _retain_centroids(numbers, thres):
    '''Only keep one number for each cluster within thres of each other'''
    numbers.sort()
    prev = -1
    ret = []
    for n in numbers:
        if prev < 0 or n - prev > thres:
            ret.append(n)
            prev = n
    return ret

def _split_vlines_hlines(lines):
    '''Separates lines into horizontal and vertical ones'''
    vlines, hlines = [], []
    for line in lines:
        (vlines if line.x1 - line.x0 < 0.1 else hlines).append(line)
    return vlines, hlines

def _npiter(arr):
    '''Wrapper for iterating numpy array'''
    for a in np.nditer(arr, flags=['refs_ok']):
        c = a.item()
        if c is not None:
            yield c




'''
Created on Oct 11, 2015

@author: xiao
'''
import os
from sys import platform as _platform

import numpy as np
from PIL import ImageFont, Image, ImageDraw

from pdftotree.pdf.vector_utils import center
from pdfminer.layout import LTAnno

white = (255, 255, 255)
black = (0, 0, 0)
red = (255, 0, 0)
green = (0, 255, 0)
blue = (0, 0, 255)
default_font_size = 10
_font_cache = {}


def lazy_load_font(font_size=default_font_size):
    '''
    Lazy loading font according to system platform
    '''
    if font_size not in _font_cache:
        if _platform.startswith('darwin'):
            font_path = "/Library/Fonts/Arial.ttf"
        elif _platform.startswith('linux'):
            font_path = "/usr/share/fonts/truetype/ubuntu-font-family/UbuntuMono-R.ttf"
        elif _platform.startswith('win32'):
            font_path = "C:\\Windows\\Fonts\\arial.ttf"
        _font_cache[font_size] = ImageFont.truetype(font_path, font_size)
    return _font_cache[font_size]


def normalize_bbox(coords, ymax, scaler=2):
    '''
    scales all coordinates and flip y axis due to different
    origin coordinates (top left vs. bottom left)
    '''
    return [coords[0] * scaler,
            ymax - (coords[3] * scaler),
            coords[2] * scaler,
            ymax - (coords[1] * scaler)]


def normalize_pts(pts, ymax, scaler=2):
    '''
    scales all coordinates and flip y axis due to different
    origin coordinates (top left vs. bottom left)
    '''
    return [(x * scaler, ymax - (y * scaler)) for x, y in pts]


def create_img(bbox=(0, 0, 200, 200)):
    # create new white image
    img = Image.new("RGBA", bbox[-2:], white)
    return img, ImageDraw.Draw(img)


def render(draw, bbox, text=None, align=None, fill=None, outline=black):
    draw.rectangle(bbox, outline=outline, fill=fill)
    if text:
        coord = center(bbox) if align == 'center' else bbox[:2]
        draw.text(coord, text, black, font=lazy_load_font())


def load_image(pdf_path, page_num):
    pdf_file = os.path.basename(pdf_path)
    basename = pdf_file[:-4]
    image_name = '%s-%06d.png' % (basename, page_num + 1)
    scan = Image.open(os.path.join('private/imgs/', image_name))
    #     scan.filter(ImageFilter.GaussianBlur(2)).show()
    # make it black and white for simplicity
    return scan.convert('1')


def load_pixels(pdf_path, page_num):
    scan_img = load_image(pdf_path, page_num)
    raw_data = np.array(scan_img.getdata())
    return (raw_data > 0).reshape(scan_img.height, scan_img.width), scan_img


def fill(mat, orig_bbox, margin):
    pass


def render_debug_img(file_name,
                     page_num,
                     elems,
                     nodes=[],
                     scaler=1,
                     print_segments=False,
                     print_curves=True,
                     print_table_bbox=True,
                     print_text_as_rect=True,
                     ):
    '''
    Shows an image rendering of the pdf page along with debugging
    info printed
    '''
    # For debugging show the boolean pixels in black white grayscale
    height = scaler * int(elems.layout.height)
    width = scaler * int(elems.layout.width)
    debug_img, draw = create_img((0, 0, width, height))
    font = lazy_load_font()
    large_font = lazy_load_font(24)

    if print_curves:
        for i, c in enumerate(elems.curves):
            if len(c.pts) > 1:
                draw.polygon(c.pts, outline=blue)
            draw.rectangle(c.bbox, fill=blue)
            # for fig in elems.figures:
            #     draw.rectangle(fig.bbox, fill = blue)

    for i, m in enumerate(elems.mentions):
        if isinstance(m, LTAnno): continue
        if print_text_as_rect:
            fill = 'pink' if hasattr(m, 'feats') and m.feats['is_cell'] else green
            #             fill = green
            draw.rectangle(m.bbox, fill=fill)
            # draw.text(center(m.bbox), str(i), black, font = font) # Draw id
            draw.text(m.bbox[:2], m.get_text(), black, font=font)  # Draw mention content
        else:
            draw.text(m.bbox[:2], m.get_text(), 'black', font=font)

    if print_segments:
        # draw skeleton for all segments
        for i, s in enumerate(elems.segments):
            draw.line(s.bbox, fill='black')

    if print_table_bbox:
        for node in nodes:
            is_table = node.is_table()
            color = 'red' if is_table else 'green'
            draw.rectangle(node.bbox, outline=color)
            if is_table:
                # text = 'Borderless' if node.is_borderless() else 'Bordered'
                text = 'Table'
                draw.rectangle(node.bbox, outline=color)
                draw.text(node.bbox[:2], text, red, font=large_font)

    # Water mark with file name so we can identify among multiple images
    if file_name and page_num is not None:
        water_mark = file_name + ':page ' + str(page_num + 1) + '@%dx%d' % (width, height)
        draw.text((10, 10), water_mark, black, font=font)
    debug_img.show()
    return debug_img



'''
Created on Jan 25, 2016

@author: xiao
'''
import collections
import logging
import numpy as np
from itertools import chain
from pdfminer.layout import LTTextLine, LTChar, LTAnno, LTCurve, LTComponent, LTLine
from pdftotree.pdf.vector_utils import *

log = logging.getLogger(__name__)

def traverse_layout(root, callback):
    '''
    Tree walker and invokes the callback as it
    traverse pdf object tree
    '''
    callback(root)
    if isinstance(root, collections.Iterable):
        for child in root:
            traverse_layout(child, callback)


def get_near_items(tree,tree_key):
    '''
    Check both possible neighbors for key
    in a binary tree
    '''
    try:
        yield tree.floor_item(tree_key)
    except KeyError:
        pass
    try:
        yield tree.ceiling_item(tree_key)
    except KeyError:
        pass

def align_add(tree, key, item, align_thres = 2.0):
    '''
    Adding the item object to a binary tree with the given
    key while allow for small key differences
    close_enough_func that checks if two keys are
    within threshold
    '''
    for near_key, near_list in get_near_items(tree, key):
        if abs(key - near_key) < align_thres:
            near_list.append(item)
            return
    # Create a new group if no items are close
    tree[key] = [item]


right_wall = lambda m: (m.x1,m.y0,m.x1,m.y1)
left_wall = lambda m: (m.x0,m.y0,m.x0,m.y1)
top_wall = lambda m: (m.x0,m.y0,m.x1,m.y0)
bot_wall = lambda m: (m.x0,m.y1,m.x1,m.y1)
def vlines_between(plane, prev, m):
    if not prev or not m: return []
    if prev.xc > m.xc: prev, m = m, prev
    query = (prev.xc, prev.yc, m.xc, prev.yc)
    return [l for l in plane.find(query) if l.x1 - l.x0 < 0.1]

def hlines_between(plane, prev, m):
    if not prev or not m: return []
    if prev.yc > m.yc: prev, m = m, prev
    query = (prev.xc, prev.yc, prev.xc, m.yc)
    return [l for l in plane.find(query) if l.y1 - l.y0 < 0.1]


def is_same_row(m1,m2):
    # Corner case for row separation
    #------
    #-prev- ------
    #------ ---m--
    #       ------
    return m1 and m2 and m2.yc > m1.y0 and m2.yc < m1.y1

def is_vline(l):
    return l.x1 - l.x0 < 0.1

def is_hline(l):
    return l.y1 - l.y0 < 0.1
#
# def align_add_to_tree(tree, key, item, close_enough_func):
#     '''
#     Adding the item object to a binary tree with the given
#     key while allow for small key differences
#     close_enough_func that checks if two keys are
#     within threshold
#     '''
#     has_neighbor = False
#     for near_key, near_list in get_near_items(tree, key):
#         if close_enough_func(key, near_key):
#             near_list.append(item)
#             has_neighbor = True
#             break
#
#     # Create a new group if no items are close
#     if not has_neighbor:
#         tree[key] = [item]
#

def collect_table_content(table_bboxes,elems):
    '''
    Returns a list of elements that are contained inside
    the corresponding supplied bbox.
    '''
    # list of table content chars
    table_contents = [[] for _ in xrange(len(table_bboxes))]
    prev_content = None
    prev_bbox = None
    for cid, c in enumerate(elems):
        # Annotations should not fall outside alone
        if isinstance(c, LTAnno):
            if prev_content is not None:
                prev_content.append(c)
            continue
        # Generally speaking table contents should be included sequentially
        # and we can avoid checking all tables for elems inside
        # Elements only need to intersect the bbox for table as some
        # formatting of fonts may result in slightly out of bbox text
        if prev_bbox is not None and intersect(prev_bbox,c.bbox):
            prev_content.append(c)
            continue
        # Search the rest of the tables for membership when done with
        # the current one
        for table_id,table_bbox in enumerate(table_bboxes):
            if intersect(table_bbox, c.bbox):
                prev_bbox = table_bbox
                prev_content = table_contents[table_id]
                prev_content.append(c)
                break
    return table_contents

_bbox = namedtuple('_bbox', ['bbox'])
_inf_bbox = _bbox([float('inf')] * 4)

def _gaps_from(intervals):
    '''
    From a list of intervals extract
    a list of sorted gaps in the form of [(g,i)]
    where g is the size of the ith gap.
    '''
    sliding_window = izip(intervals, intervals[1:])
    gaps = [b[0] - a[1] for a, b in sliding_window]
    return gaps

def project_onto(objs, axis, min_gap_size = 4.0):
    '''
    Projects object bboxes onto the axis and return the
    unioned intervals and groups of objects in intervals.
    '''
    if axis == 'x': axis = 0
    if axis == 'y': axis = 1
    axis_end = axis + 2
    if axis == 0: # if projecting onto X axis
        objs.sort(key = lambda o:o.x0)
    else:
        objs.sort(key = lambda o:o.y0)

    intervals = []
    groups = []

    start_i = 0
    start = objs[0].bbox[axis]
    end = objs[0].bbox[axis_end]

    # Use _inf_bbox to trigger the last interval divide
    for o_i, o in enumerate(chain(objs,[_inf_bbox])):

        # Get current interval
        o_start = o.bbox[axis]
        o_end = o.bbox[axis_end]

        # start new interval when gap with previous end is big
        if o_start > end + min_gap_size:

            # Append new interval coordinates for children
            intervals.append((start, end))

            # Append child object group on page
            groups.append(objs[start_i:o_i])

            # Mark next obj list range
            start_i = o_i
            start = o_start

        # Always check to extend current interval to new end
        if o_end > end:
            end = o_end
        # else do nothing
    return intervals, groups

def recursive_xy_divide(elems, avg_font_size):
    '''
    Recursively group/divide the document by white stripes
    by projecting elements onto alternating axes as intervals.

    avg_font_size: the minimum gap size between elements below
    which we consider interval continuous.
    '''
    log.info(avg_font_size)
    objects = list(elems.mentions)
    objects.extend(elems.segments)
    bboxes = []
    # A tree that is a list of its children
    # bboxes can be recursively reconstructed from
    # the leaves
    def divide(objs, bbox, h_split = True, is_single = False):
        '''
        Recursive wrapper for splitting a list of objects
        with bounding boxes.
        h_split: whether to split along y axis, otherwise
        we split along x axis.
        '''
        if not objs: return []

        # range start/end indices
        axis = 1 if h_split else 0

        intervals, groups = project_onto(objs, axis, avg_font_size)

        # base case where we can not actually divide
        single_child = len(groups) == 1

        # Can not divide in both X and Y, stop
        if is_single and single_child:
            bboxes.append(bbox)
            return objs
        else:
            children = []

            for interval, group in izip(intervals, groups):
                # Create the bbox for the subgroup
                sub_bbox = np.array(bbox)
                sub_bbox[[axis, axis + 2]] = interval
                # Append the sub-document tree
                child = divide(group, sub_bbox, not h_split, single_child)
                children.append(child)
            return children

    full_page_bbox = (0, 0, elems.layout.width, elems.layout.height)
    # Filter out invalid objects
    objects = [o for o in objects if inside(full_page_bbox,o.bbox)]
    log.info('avg_font_size for dividing', avg_font_size)
    tree = divide(objects, full_page_bbox) if objects else []
    return bboxes, tree




'''
This script takes a PDF document and extracts it's tree structure and then
writes the HTML based on that tree structure. The components of the tree
structure are:
- Tables
- Table Captions
- Figures
- Figure Captions
- Section Headers
- Paragraphs
- List (References in research papers)
- Page Headers

Tables are detected using a Machine learning model, provide the path in
model_path argument = TreeStructure/data/paleo/ml/model.pkl.

Other tree parts are detected using heuristic methods.

Set favor_figures to "False" for Hardware sheets.
'''
from pdftotree.TreeExtract import TreeExtractor
from pdftotree.TreeVisualizer import TreeVisualizer
import codecs
import logging
import os
import pickle
import re


def load_model(model_path):
    log = logging.getLogger(__name__)
    log.info("Loading pretrained model for table detection")
    model = pickle.load(open(model_path, 'rb'))
    log.info("Model loaded!")
    return model


def visualize_tree(pdf_file, pdf_tree, html_path):
    v = TreeVisualizer(pdf_file)
    filename_prefix = os.path.basename(pdf_file)
    v.display_candidates(pdf_tree, html_path, filename_prefix)


def parse(pdf_file,
          html_path=None,
          model_path=None,
          favor_figures=True,
          visualize=False):

    log = logging.getLogger(__name__)

    model = None
    if (model_path is not None):
        model = load_model(model_path)
    extractor = TreeExtractor(pdf_file)
    if (not extractor.is_scanned()):
        log.info("Digitized PDF detected, building tree structure...")
        pdf_tree = extractor.get_tree_structure(model, favor_figures)
        log.info("Tree structure built, creating html...")
        pdf_html = extractor.get_html_tree()
        log.info("HTML created, outputting...")
        pdf_filename = os.path.basename(pdf_file)
        # Check html_path exists, create if not
        pdf_html = re.sub(r'[\x00-\x1F]+', '', pdf_html)

        if html_path is None:
            return pdf_html
        elif not os.path.exists(html_path):
            os.makedirs(html_path)
        with codecs.open(
                html_path + pdf_filename[:-4] + ".html",
                encoding="utf-8",
                mode="w") as f:
            f.write(pdf_html)
        if visualize:
            visualize_tree(pdf_file, pdf_tree, html_path)
    else:
        log.error("Document is scanned, cannot build tree structure")



from __future__ import division
from builtins import str
import string

from pdftotree.pdf.pdf_parsers import *
from pdftotree.pdf.vector_utils import *
from pdftotree.utils.bbox_utils import isContained


# ******************* Table Coverage Features *************************************


def get_area_coverage(bbox):
    b = bbox[-4:]
    return ((b[2] - b[0]) * (b[3] - b[1])) / float(bbox[1] * bbox[2])


def get_width_coverage(bbox):
    b = bbox[-4:]
    return (b[3] - b[1]) / float(bbox[1])


def get_height_coverage(bbox):
    b = bbox[-4:]
    return (b[2] - b[0]) / float(bbox[2])


# ******************* Text Coverage Features *************************************

def get_mentions_within_bbox(bbox, mentions):
    mentions_within_bbox = []
    for mention in mentions:
        bbox_mention = (int(mention.y0), int(mention.x0), int(mention.y1), int(mention.x1))
        if isContained(bbox_mention, bbox[-4:]):
            mentions_within_bbox += [mention]
    return mentions_within_bbox


def get_text_sparsity(bbox, mentions):
    """
    computes text area coverage
    :param mentions:
    :return:
    """
    b = bbox[-4:]
    bbox_area = ((b[2] - b[0]) * (b[3] - b[1]))
    text_area = 0
    for m in mentions:
        text_area += (m.x1 - m.x0) * (m.y1 - m.y0)
    try:
        return 1.0 * text_area / bbox_area
    except ZeroDivisionError:
        return 0.


def symbols_and_numbers_density(bbox, mentions):
    symbols = set(string.punctuation)
    num_symbols = sum([1 for elem in mentions if elem.get_text() in symbols])
    num_numbers = 0
    for elem in mentions:
        num_numbers += sum([c.isdigit() for c in elem.get_text()])
    return [num_symbols, num_numbers]


# ******************* Lines Features *************************************

def get_lines_within_bbox(bbox, segments):
    lines_within_bbox = []
    for line in segments:
        bbox_line = (int(line.y0), int(line.x0), int(line.y1), int(line.x1))
        if isContained(bbox_line, bbox[-4:]):
            lines_within_bbox += [line]
    return lines_within_bbox


def get_lines_features(bboxes, elems):
    features = []
    for bbox in bboxes:
        mentions = get_mentions_within_bbox(bbox, elems.mentions)
        segments = get_lines_within_bbox(bbox, elems.segments)
        feat = [get_area_coverage(bbox)]
        feat += [get_height_coverage(bbox)]
        feat += [get_width_coverage(bbox)]
        feat += [get_text_sparsity(bbox, mentions)]
        feat += symbols_and_numbers_density(bbox, mentions)
        feat += [len(segments)]
        features += [feat]
    return features

# ******************* Alignments Features *************************************


def get_alignment_features(line_bboxes, elems, font_stat):
    alignment_features = []
    for line_bbox in line_bboxes:
        line_bbox_ordered = (line_bbox[4], line_bbox[3], line_bbox[6], line_bbox[5])
        boxes = [elem for elem in elems.mentions if intersect(line_bbox_ordered, elem.bbox)]
        boxes_segments = [elem for elem in elems.segments if intersect(line_bbox_ordered, elem.bbox)]
        boxes_figures = [elem for elem in elems.figures if intersect(line_bbox_ordered, elem.bbox)]
        boxes_curves = [elem for elem in elems.curves if intersect(line_bbox_ordered, elem.bbox)]
        page_width = elems.layout.width
        page_height = elems.layout.height
        avg_font_pts = get_most_common_font_pts(elems.mentions, font_stat)
        width = get_page_width(boxes + boxes_segments + boxes_figures + boxes_curves)
        if (len(boxes) == 0):
            alignment_features += [[0] * 17]
            continue
        char_width = get_char_width(boxes)
        grid_size = avg_font_pts / 2.0
        for i, m in enumerate(boxes + elems.figures):
            m.id = i
            m.feats = defaultdict(bool)
            prefix = ''
            if isinstance(m, LTTextLine) and m.font_name: prefix = m.font_name + '-' + str(m.font_size) + '-'
            m.xc = (m.x0 + m.x1) / 2.0
            m.yc = (m.y0 + m.y1) / 2.0
            m.feats[prefix + 'x0'] = m.x0_grid = m.x0 // grid_size
            m.feats[prefix + 'x1'] = m.x1_grid = m.x1 // grid_size
            m.feats[prefix + 'xc'] = m.xc_grid = m.xc // grid_size
            m.feats[prefix + 'yc'] = m.yc_grid = m.yc // grid_size

        nodes, nodes_features = cluster_vertically_aligned_boxes(boxes, elems.layout.bbox, avg_font_pts, width,
                                                                 char_width, boxes_segments, boxes_curves,
                                                                 boxes_figures, page_width, True)
        if (len(nodes_features) == 0):
            alignment_features += [[0] * 17]
        else:
            alignment_features += [nodes_features]
    return alignment_features



'''
Handles abstract rendering of the layout
in order to extract local visual features

Created on Jan 28, 2016

@author: xiao
'''
from pdf.vector_utils import *
import logging
import numpy as np

class Renderer(object):
    '''
    enumeration objects to be placed into the
    rendered image
    '''
    empty = 0
    horizontal_line = -1
    vertical_line = -2
    text = -3
    img = -4
    curve = -5
    misc = -6

    def __init__(self, elems, scaler = 1):
        '''
        Initializes the rendered object grid with specified
        scaler so we can map original coordinates into the
        new grid map.
        '''
        self.log = logging.getLogger(__name__)
        self.scaler = scaler
        layout = elems.layout
        width = int(np.ceil(scaler * layout.width))
        height = int(np.ceil(scaler * layout.height))

        self.grid = np.zeros((width, height), dtype = np.int8)

        # Estimates the grid size in megabytes
        self.log.info(self.grid.nbytes/float(1048576))
        for line in elems.segments:
            if line.height < 0.1:  # Horizontal lines
                self.draw_rect(line.bbox, self.horizontal_line)
            elif line.width < 0.1:# Vertical lines
                self.draw_rect(line.bbox, self.vertical_line)

        for mention in elems.mentions:
            self.draw_rect(mention.bbox, self.text)

        for figure in elems.figures:
            self.draw_rect(figure.bbox, self.img)

    def draw_rect(self, bbox, cell_val):
        '''
        Fills the bbox with the content values
        Float bbox values are normalized to have non-zero area
        '''
        new_x0 = int(bbox[x0])
        new_y0 = int(bbox[y0])
        new_x1 = max(new_x0 + 1, int(bbox[x1]))
        new_y1 = max(new_y0 + 1, int(bbox[y1]))

        self.grid[new_x0:new_x1,new_y0:new_y1] = cell_val

    @staticmethod
    def is_mention(cell_val):
        '''
        Nonnegative values in grid cells are reserved for mention ids
        '''
        return cell_val >= 0




TOLERANCE = 5


def doOverlap(bbox1, bbox2):
    """
    :param bbox1: bounding box of the first rectangle
    :param bbox2: bounding box of the second rectangle
    :return: 1 if the two rectangles overlap
    """
    if bbox1[2] < bbox2[0] or bbox2[2] < bbox1[0]:
        return False
    if bbox1[3] < bbox2[1] or bbox2[3] < bbox1[1]:
        return False
    return True


def isContained(bbox1, bbox2, tol=TOLERANCE):
    """
    :param bbox1: bounding box of the first rectangle
    :param bbox2: bounding box of the second rectangle
    :return: True if bbox1 is contaned in bbox2
    """
    if bbox1[0] > bbox2[0] - tol and bbox1[1] > bbox2[1] - tol:
        if bbox1[2] < bbox2[2] + tol and bbox1[3] < bbox2[3] + tol:
            return True
    return False


def mergeBboxes(bbox1, bbox2):
    """
    :param bbox1: (top, left, bottom, right)
    :param bbox2: (top, left, bottom, right)
    :return: Merge bounding boxes
    """
    if isContained(bbox1, bbox2):
        return bbox2
    elif isContained(bbox2, bbox1):
        return bbox1
    else:
        return (min(bbox1[0], bbox2[0]), min(bbox1[1], bbox2[1]),
                max(bbox1[2], bbox2[2]), max(bbox1[3], bbox2[3]))


def get_rectangles(vertical_lines, horizontal_lines):
    """
    :param vertical_lines: list of vertical lines coordinates
    :param horizontal_lines: list of horizontal lines coordinates
    :return: List of bounding boxes for tables
    """
    rectangles = []
    i = 0
    j = 0
    while i < len(horizontal_lines) and j < len(vertical_lines):
        if int(horizontal_lines[i][0]) == vertical_lines[j][0]:
            if int(horizontal_lines[i][1]) == int(vertical_lines[j][1]):
                h = horizontal_lines[i]
                v = vertical_lines[j]
                rectangles += [(v[0], h[1], v[2], h[3])]
                i += 1
                j += 1
            elif int(horizontal_lines[i][1]) < int(vertical_lines[j][1]):
                i += 1
            else:
                j += 1
        elif int(horizontal_lines[i][0]) < int(vertical_lines[j][0]):
            i += 1
        else:
            j += 1
    rectangles = [
        r for r in rectangles
        if ((r[2] - r[0]) > TOLERANCE and (r[3] - r[1]) > TOLERANCE)
    ]
    return rectangles


def get_outer_bounding_boxes(rectangles):
    """
    :param rectangles: list of bounding boxes (top, left, bottom, right)
    :return: outer bounding boxes (only the largest bbox when bboxes intersect)
    """
    if len(rectangles) == 0:
        return []
    outer_bboxes = [rectangles[0]]
    for bbox2 in rectangles[1:]:
        overlap_indexes = []
        for i, bbox1 in enumerate(outer_bboxes):  # TODO: optimize this !!
            if doOverlap(bbox1, bbox2):
                overlap_indexes.append(i)
        for i in overlap_indexes:
            bbox2 = mergeBboxes(bbox2, outer_bboxes[i])
        for i in sorted(overlap_indexes, reverse=True):
            del outer_bboxes[i]
        outer_bboxes.append(bbox2)
    return outer_bboxes


def get_intersection(bbox1, bbox2):
    """
    :param bbox1: (page, width, height, top, left, bottom, right)
    :param bbox2: (page, width, height, top, left, bottom, right)
    :return: intersection if bboxes are in the same page and intersect
    """
    intersection = []
    page_1, page_width, page_height, top_1, left_1, bottom_1, right_1 = bbox1
    page_2, _, _, top_2, left_2, bottom_2, right_2 = bbox2
    if page_1 == page_2:
        if doOverlap((top_1, left_1, bottom_1, right_1),
                     (top_2, left_2, bottom_2, right_2)):
            intersection += [(page_1, page_width, page_height, max(
                top_1, top_2), max(left_1, left_2), min(bottom_1, bottom_2),
                              min(right_1, right_2))]
    return intersection


def compute_iou(bbox1, bbox2):
    """
    :param bbox1: (page, width, height, top, left, bottom, right)
    :param bbox2: (page, width, height, top, left, bottom, right)
    :return: intersection over union if bboxes are in the same page and intersect
    """
    top_1, left_1, bottom_1, right_1 = bbox1
    top_2, left_2, bottom_2, right_2 = bbox2
    if doOverlap((top_1, left_1, bottom_1, right_1),
                 (top_2, left_2, bottom_2, right_2)):
        intersection = (min(bottom_1, bottom_2) - max(top_1, top_2)) * (
            min(right_1, right_2) - max(left_1, left_2))
        union = (bottom_1 - top_1) * (right_1 - left_1) + (
            bottom_2 - top_2) * (right_2 - left_2) - intersection
        return float(intersection) / float(union)
    return 0.



import numpy as np
from wand.color import Color
from wand.display import display
from wand.drawing import Drawing
from wand.image import Image


def display_bounding_boxes(img, blocks, alternatecolors=False, color=Color('blue')):
    """
    Displays each of the bounding boxes passed in 'boxes' on an image of the pdf
    pointed to by pdf_file
    boxes is a list of 5-tuples (page, top, left, bottom, right)
    """
    draw = Drawing()
    draw.fill_color = Color('rgba(0, 0, 0, 0)')
    draw.stroke_color = color
    for block in blocks:
        top, left, bottom, right = block[-4:]
        if alternatecolors:
            draw.stroke_color = Color('rgba({},{},{}, 1)'.format(
                    str(np.random.randint(255)), str(np.random.randint(255)), str(np.random.randint(255))))
        draw.rectangle(left=float(left), top=float(top), right=float(right), bottom=float(bottom))
        draw(img)
    display(img)


def display_bounding_boxes_within_notebook(page_num, extractor, blocks, alternatecolors=False, color=Color('blue')):
    """
    Displays each of the bounding boxes passed in 'boxes' on an image of the pdf
    pointed to by pdf_file
    boxes is a list of 5-tuples (page, top, left, bottom, right)
    """
    elems = extractor.elems[page_num]
    page_width, page_height = int(elems.layout.width), int(elems.layout.height)
    img = pdf_to_img(extractor.pdf_file, page_num, page_width, page_height)
    draw = Drawing()
    draw.fill_color = Color('rgba(0, 0, 0, 0)')
    draw.stroke_color = color
    for block in blocks:
        top, left, bottom, right = block[-4:]
        if alternatecolors:
            draw.stroke_color = Color('rgba({},{},{}, 1)'.format(
                    str(np.random.randint(255)), str(np.random.randint(255)), str(np.random.randint(255))))
        draw.rectangle(left=float(left), top=float(top), right=float(right), bottom=float(bottom))
        draw(img)
    return img


def pdf_to_img(pdf_file, page_num, page_width, page_height):
    """
    Converts pdf file into image
    :param pdf_file: path to the pdf file
    :param page_num: page number to convert (index starting at 1)
    :return: wand image object
    """
    img = Image(filename='{}[{}]'.format(pdf_file, page_num - 1))
    img.resize(page_width, page_height)
    return img





import os
import subprocess

from collections import defaultdict
from wand.image import Image
from wand.drawing import Drawing
from wand.color import Color
from bs4 import BeautifulSoup
from IPython.display import display
from PIL import ImageFont


class TreeVisualizer():
    """
    Object to display bounding boxes on a pdf document
    """

    def __init__(self, pdf_file):
        """
        :param pdf_path: directory where documents are stored
        :return:
        """
        self.pdf_file = pdf_file

    def display_boxes(self, tree, html_path, filename_prefix, alternate_colors=False):
        """
        Displays each of the bounding boxes passed in 'boxes' on images of the pdf
        pointed to by pdf_file
        boxes is a list of 5-tuples (page, top, left, bottom, right)
        """
        imgs = []
        colors = {"section_header":Color('blue'), "figure":Color('green'), "figure_caption":Color('green'), "table_caption":Color('red'), "list":Color('yellow'), "paragraph":Color('gray'), "table":Color('red'), "header":Color("brown")}
        for i, page_num in enumerate(tree.keys()):
            img = self.pdf_to_img(page_num)
            draw = Drawing()
            draw.fill_color = Color('rgba(0, 0, 0, 0.0)')
            for clust in tree[page_num]:
                for (pnum, pwidth, pheight, top, left, bottom, right) in tree[page_num][clust]:
                    draw.stroke_color = colors[clust]
                    draw.rectangle(left=left, top=top, right=right, bottom=bottom)
                    draw.push()
                    draw.font_size = 20
                    draw.font_weight = 10
                    draw.fill_color =  colors[clust]
                    if int(left)>0 and int(top)>0:
                        draw.text(x=int(left), y=int(top), body=clust)
                    draw.pop()
            draw(img)
            img.save(filename=html_path + filename_prefix + "_page_" + str(i) + '.png')
            imgs.append(img)
        return imgs

    def display_candidates(self, tree, html_path, filename_prefix):
        """
        Displays the bounding boxes corresponding to candidates on an image of the pdf
        boxes is a list of 5-tuples (page, top, left, bottom, right)
        """
        imgs = self.display_boxes(tree, html_path, filename_prefix, alternate_colors=True)
        return display(*imgs)

    def pdf_to_img(self, page_num, pdf_dim=None):
        """
        Converts pdf file into image
        :param pdf_file: path to the pdf file
        :param page_num: page number to convert (index starting at 1)
        :return: wand image object
        """
        if not pdf_dim:
            pdf_dim = get_pdf_dim(self.pdf_file)
        page_width, page_height = pdf_dim
        img = Image(filename='{}[{}]'.format(self.pdf_file, page_num-1))
        img.resize(page_width, page_height)
        return img

def get_pdf_dim(pdf_file):
    html_content = subprocess.check_output(
        "pdftotext -f {} -l {} -bbox '{}' -".format('1', '1', pdf_file), shell=True)
    soup = BeautifulSoup(html_content, "html.parser")
    pages = soup.find_all('page')
    page_width, page_height = int(float(pages[0].get('width'))), int(float(pages[0].get('height')))
    return page_width, page_height





'''
Created on Oct 26, 2015
Parsing raw PDF data into python data structures

@author: xiao
'''
from __future__ import division
from builtins import filter
from builtins import str
from builtins import zip
from builtins import range
import logging
import math
import operator
from collections import defaultdict
from copy import deepcopy
from functools import cmp_to_key
from pdfminer.utils import Plane
from pdftotree.pdf.layout_utils import *
from pdftotree.pdf.node import Node

log = logging.getLogger(__name__)


def parse_layout(elems, font_stat, combine=False):
    '''
    Parses pdf texts into a hypergraph grouped into rows
    and columns and then output
    '''
    boxes_segments = elems.segments
    boxes_curves = elems.curves
    boxes_figures = elems.figures
    page_width = elems.layout.width
    page_height = elems.layout.height
    boxes = elems.mentions
    avg_font_pts = get_most_common_font_pts(elems.mentions, font_stat)
    width = get_page_width(
        boxes + boxes_segments + boxes_figures + boxes_curves)
    char_width = get_char_width(boxes)
    grid_size = avg_font_pts / 2.0
    for i, m in enumerate(boxes + elems.figures):
        m.id = i
        m.feats = defaultdict(bool)
        prefix = ''
        if isinstance(m, LTTextLine) and m.font_name:
            prefix = m.font_name + '-' + str(m.font_size) + '-'
        m.xc = (m.x0 + m.x1) / 2.0
        m.yc = (m.y0 + m.y1) / 2.0
        m.feats[prefix + 'x0'] = m.x0_grid = m.x0 // grid_size
        m.feats[prefix + 'x1'] = m.x1_grid = m.x1 // grid_size
        m.feats[prefix + 'xc'] = m.xc_grid = m.xc // grid_size
        m.feats[prefix + 'yc'] = m.yc_grid = m.yc // grid_size

    tbls, tbl_features = cluster_vertically_aligned_boxes(
        boxes, elems.layout.bbox, avg_font_pts, width, char_width,
        boxes_segments, boxes_curves, boxes_figures, page_width, combine)
    return tbls, tbl_features


def cluster_vertically_aligned_boxes(boxes, page_bbox, avg_font_pts, width,
                                     char_width, boxes_segments, boxes_curves,
                                     boxes_figures, page_width, combine):
    # Too many "." in the Table of Content pages
    if (len(boxes) == 0 or len(boxes) > 3500):
        return []
    plane = Plane(page_bbox)
    plane.extend(boxes)

    # initialize clusters
    cid2obj = [set([i]) for i in range(len(boxes))]
    # default object map to cluster with its own index
    obj2cid = list(range(len(boxes)))
    prev_clusters = obj2cid
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2cid[i1] == obj2cid[i2])):
                    continue
                if (b1.bbox[1] < b2.bbox[1]):
                    box1 = b1.bbox
                    box2 = b2.bbox
                elif (b2.bbox[1] < b1.bbox[1]):
                    box1 = b2.bbox
                    box2 = b1.bbox
                else:
                    # horizontally aligned
                    continue
                if (box2[1] < box1[3]
                        or (box2[1] - box1[1] < 1.5 * avg_font_pts)
                        or (box2[3] - box1[3] < 1.5 * avg_font_pts)):
                    # can probably do better if we find the average space
                    # between words
                    if (abs(box1[0] - box2[0]) < 3
                            or abs(box1[2] - box2[2]) < 3
                            or (((box1[0] + box1[2]) / 2) == (
                                (box2[0] + box2[2]) / 2))
                            or ((box1[0] < box2[0]) and (box1[2] > box2[0]))
                            or ((box1[0] > box2[0]) and (box2[2] > box1[0]))):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        # move all objects from cluster cid2 to cid1
                        # reassign cluster ids for all such objects as well
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
        if (prev_clusters == obj2cid):
            break
        prev_clusters = obj2cid
    clusters = [[boxes[i] for i in cluster]
                for cluster in filter(bool, cid2obj)]

    rid2obj = [set([i]) for i in range(len(boxes))]  # initialize clusters
    # default object map to cluster with its own index
    obj2rid = list(range(len(boxes)))
    prev_clusters = obj2rid
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2rid[i1] == obj2rid[i2])):
                    continue
                box1 = b1.bbox
                box2 = b2.bbox
                if ((abs(box1[1] - box2[1]) < 0.11 * avg_font_pts)
                        or ((abs(box1[3] - box2[3]) < 0.11 * avg_font_pts))
                        or (round((box1[1] + box1[3]) / 2) == round(
                            (box2[1] + box2[3]) / 2))):
                    min_i = min(i1, i2)
                    max_i = max(i1, i2)
                    rid1 = obj2rid[min_i]
                    rid2 = obj2rid[max_i]
                    for obj_iter in rid2obj[rid2]:
                        rid2obj[rid1].add(obj_iter)
                        obj2rid[obj_iter] = rid1
                    rid2obj[rid2] = set()
        if (prev_clusters == obj2rid):
            break
        prev_clusters = obj2rid

    not_merge = set()
    for i1, b1 in enumerate(boxes):
        for i2 in cid2obj[obj2cid[i1]]:
            if (i1 == i2):
                continue
            row1 = obj2rid[i1]
            row2 = obj2rid[i2]
            if (row1 == row2):
                continue
            if (b1.bbox[1] < b2.bbox[1]):
                box1 = b1.bbox
                box2 = b2.bbox
            elif (b2.bbox[1] < b1.bbox[1]):
                box1 = b2.bbox
                box2 = b1.bbox
            else:
                # horizontally aligned
                continue
            text_1 = 0.0
            for obj in rid2obj[row1]:
                text_1 += boxes[obj].bbox[2] - boxes[obj].bbox[0]
            text_2 = 0.0
            for obj in rid2obj[row2]:
                text_2 += boxes[obj].bbox[2] - boxes[obj].bbox[0]
            if (abs(text_1 - text_2) / width > 0.1):
                min_i = min(i1, i2)
                max_i = max(i1, i2)
                not_merge.add((min_i, max_i))

    # Alignment Features
    # If text boxes are very close in a row
    if_row_connected = defaultdict(int)
    num_row_connected = defaultdict(lambda: 1)
    # If text is merged using span code in adjacent rows, this feature tells the number of times the cluster went through span based clustering
    if_connected_by_span = defaultdict(int)
    num_connected_by_span = defaultdict(lambda: 1)
    # If columns were merged using cluster alignment
    if_connected_by_align = defaultdict(int)
    num_connected_by_align = defaultdict(lambda: 1)
    # If vertical columns were merged
    if_vertical_columns_merged = defaultdict(int)
    num_vertical_columns_merged = defaultdict(lambda: 1)
    # Number of Line Segments, Curves and Figures
    num_segments = defaultdict(int)
    num_curves = defaultdict(int)
    num_figures = defaultdict(int)
    # Average Word Space
    total_word_space = defaultdict(float)
    avg_word_space = defaultdict(float)
    avg_word_space_norm = defaultdict(float)
    node_space = defaultdict(float)
    avg_node_space = defaultdict(float)
    avg_node_space_norm = defaultdict(float)

    cid2obj = [set([i]) for i in range(len(boxes))]  # initialize clusters
    obj2cid = list(range(
        len(boxes)))  # default object map to cluster with its own index
    prev_clusters = obj2cid
    # add the code for merging close text boxes in particular row
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2cid[i1] == obj2cid[i2])):
                    continue
                box1 = b1.bbox
                box2 = b2.bbox
                if (obj2rid[i1] == obj2rid[i2]):
                    if (((b1.bbox[0] < b2.bbox[0]) and
                         ((b2.bbox[0] - b1.bbox[2]) <= 2 * char_width)) or
                        ((b2.bbox[0] < b1.bbox[0]) and
                         ((b1.bbox[0] - b2.bbox[2]) <= 2 * char_width))):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
                        # Features
                        if_row_connected[cid1] = 1
                        if_row_connected[cid2] = 0
                        num_row_connected[cid1] += num_row_connected[cid2]
                        num_row_connected[cid2] = 0
        if (prev_clusters == obj2cid):
            break
        prev_clusters = obj2cid

    # vertical alignment code
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2cid[i1] == obj2cid[i2])):
                    continue
                if (b1.bbox[1] < b2.bbox[1]):
                    box1 = b1.bbox
                    box2 = b2.bbox
                elif (b2.bbox[1] < b1.bbox[1]):
                    box1 = b2.bbox
                    box2 = b1.bbox
                else:
                    # horizontally aligned
                    continue
                if (
                        box2[1] < box1[3]
                        or (box2[1] - box1[1] < 1.5 * avg_font_pts)
                        or (box2[3] - box1[3] < 1.5 * avg_font_pts)
                ):  # can probably do better if we find the average space between words
                    if (
                            abs(box1[0] - box2[0]) < 3
                            or abs(box1[2] - box2[2]) < 3
                            or (((box1[0] + box1[2]) / 2) == (
                                (box2[0] + box2[2]) / 2))
                    ):  # or ((box1[0]<box2[0]) and (box1[2]>box2[0])) or ((box1[0]>box2[0]) and (box2[2]>box1[0]))): #added center alignemnt
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        if ((min_i, max_i) not in not_merge):
                            cid1 = obj2cid[min_i]
                            cid2 = obj2cid[max_i]
                            # move all objects from cluster cid2 to cid1
                            # reassign cluster ids for all such objects as well
                            for obj_iter in cid2obj[cid2]:
                                cid2obj[cid1].add(obj_iter)
                                obj2cid[obj_iter] = cid1
                            cid2obj[cid2] = set()
                            # Features
                            if_connected_by_span[cid1] = 1
                            if_connected_by_span[cid2] = 0
                            if (if_row_connected[cid1] == 1
                                    or if_row_connected[cid2] == 1):
                                if_row_connected[cid1] = 1
                                num_row_connected[cid1] += num_row_connected[
                                    cid2]
                                num_row_connected[cid2] = 0
                                if_row_connected[cid2] = 0
                            num_connected_by_span[
                                cid1] = num_connected_by_span[cid1] + num_connected_by_span[cid2]
                            num_connected_by_span[cid2] = 0
        if (prev_clusters == obj2cid):
            break
        prev_clusters = obj2cid

    # blacklist nearly half-page wide clusters before horizontal merging
    cid2obj2 = cid2obj[:]
    obj2cid2 = obj2cid[:]
    blacklist = set()
    blacklist_obj = set()
    for cid_iter in range(len(cid2obj2)):
        cid = cid2obj2[cid_iter]
        xmin = float("Inf")
        xmax = float("-Inf")
        for obj in cid:
            xmin = min(xmin, boxes[obj].bbox[0])
            xmax = max(xmax, boxes[obj].bbox[2])
        if (((xmax - xmin) > width / 2.75 and (xmax - xmin) < width / 2)
                or ((xmax - xmin) > 0.9 * width)):
            blacklist.add(cid_iter)
            for obj in cid:
                blacklist_obj.add(obj)
                for obj_iter in rid2obj[obj2rid[obj]]:
                    if (boxes[obj_iter].bbox[0] >= xmin
                            and boxes[obj_iter].bbox[2] <= xmax):
                        blacklist_obj.add(obj_iter)

    # create a cluster span
    cid2span = {}
    for cid in range(len(cid2obj)):
        cid2span[cid] = {}
        cid2span[cid]["min_x"] = float("Inf")
        cid2span[cid]["min_y"] = float("Inf")
        cid2span[cid]["max_x"] = float("-Inf")
        cid2span[cid]["max_y"] = float("-Inf")
        for obj in cid2obj[cid]:
            cid2span[cid]["min_x"] = min(cid2span[cid]["min_x"],
                                         boxes[obj].bbox[0])
            cid2span[cid]["max_x"] = max(cid2span[cid]["max_x"],
                                         boxes[obj].bbox[2])
            cid2span[cid]["min_y"] = min(cid2span[cid]["min_y"],
                                         boxes[obj].bbox[1])
            cid2span[cid]["max_y"] = max(cid2span[cid]["max_y"],
                                         boxes[obj].bbox[3])

    cid2cid = {}
    cid_pair_compared = set()
    cid2cid2 = [cid for cid in range(len(cid2obj))]
    for i1, b1 in enumerate(boxes):
        for i2, b2 in enumerate(boxes):
            if (i1 == i2):
                continue
            if (i1 in blacklist_obj or i2 in blacklist_obj):
                continue
            cid1 = obj2cid[i1]
            cid2 = obj2cid[i2]
            if ((min(cid1, cid2), max(cid1, cid2)) in cid_pair_compared):
                continue
            if (cid1 == cid2):
                continue
            if (obj2rid[i1] == obj2rid[i2]):
                continue
            if (cid1 not in cid2cid):
                cid2cid[cid1] = set()
            if (cid2 not in cid2cid):
                cid2cid[cid2] = set()
            if (cid2span[cid1]["min_y"] < cid2span[cid2]["min_y"]):
                box1 = [
                    cid2span[cid1]["min_x"], cid2span[cid1]["min_y"],
                    cid2span[cid1]["max_x"], cid2span[cid1]["max_y"]
                ]
                box2 = [
                    cid2span[cid2]["min_x"], cid2span[cid2]["min_y"],
                    cid2span[cid2]["max_x"], cid2span[cid2]["max_y"]
                ]
            else:
                box1 = [
                    cid2span[cid2]["min_x"], cid2span[cid2]["min_y"],
                    cid2span[cid2]["max_x"], cid2span[cid2]["max_y"]
                ]
                box2 = [
                    cid2span[cid1]["min_x"], cid2span[cid1]["min_y"],
                    cid2span[cid1]["max_x"], cid2span[cid1]["max_y"]
                ]
            if (((box1[1] < box2[1]) and (box1[3] > box2[1]))
                    or ((box1[1] > box2[1]) and (box1[1] < box2[3]))):
                continue
            cid_pair_compared.add((min(cid1, cid2), max(cid1, cid2)))
            query_rect = (min(box1[0], box2[0]), min(box1[1], box2[1]),
                          max(box1[2], box2[2]), max(box1[3], box2[3]))
            connected = True
            for i3, b3 in enumerate(boxes):
                if ((i3 == i1) or (i3 == i2)):
                    continue
                if (obj2cid[i1] == obj2cid[i3] or obj2cid[i2] == obj2cid[i3]):
                    continue
                box3 = b3.bbox
                if (intersect(query_rect, box3)):
                    connected = False
                    break
            if (
                ((round(box1[0]) == round(box2[0])
                  or round(box1[2]) == round(box2[2])) and connected)
                    or (round((box1[0] + box1[2]) / 2) == round(
                        (box2[0] + box2[2]) / 2) and connected)
            ):  # or (abs((box1[0]+box1[2])/2-(box2[0]+box2[2])/2)<0.1*char_width and connected)):# or ((box1[0]<box2[0]) and (box1[2]>box2[0])) or ((box1[0]>box2[0]) and (box2[2]>box1[0]))): #added center alignemnt
                cid2cid[min(cid1, cid2)].add(max(cid1, cid2))
                min_cid = min(cid1, cid2)
                max_cid = max(cid1, cid2)
                for cid_iter in cid2cid2:
                    if (cid2cid2[cid_iter] == cid2cid2[max_cid]):
                        cid2cid2[cid_iter] = cid2cid2[min_cid]

    # post-process cid2cid
    cid2obj2 = cid2obj[:]
    obj2cid2 = obj2cid[:]
    for cid in range(len(cid2cid2)):
        cid_merge = cid2cid2[cid]
        if (cid != cid_merge):
            for obj_iter in cid2obj2[cid]:
                cid2obj2[cid_merge].add(obj_iter)
                obj2cid2[obj_iter] = cid_merge
            cid2obj2[cid] = set()
            # Features
            if_connected_by_align[cid_merge] = 1
            if_connected_by_align[cid] = 0
            if (if_row_connected[cid_merge] == 1
                    or if_row_connected[cid] == 1):
                if_row_connected[cid_merge] = 1
                num_row_connected[cid_merge] += num_row_connected[cid]
                num_row_connected[cid] = 0
                if_row_connected[cid2] = 0
            if (if_connected_by_span[cid_merge] == 1
                    or if_connected_by_span[cid] == 1):
                if_connected_by_span[cid_merge] = 1
                num_connected_by_span[cid_merge] += num_connected_by_span[cid]
                num_connected_by_span[cid] = 0
                if_connected_by_span[cid] = 0
            num_connected_by_align[cid_merge] += num_connected_by_align[cid]
            num_connected_by_align[cid] = 0

            # code to merge columns for table
    prev_clusters = obj2cid2
    while (True):
        for obj1, b1 in enumerate(boxes):
            cid1 = obj2cid2[obj1]
            rid1 = obj2rid[obj1]
            if (cid1 in blacklist):
                continue
            if (obj1 in blacklist_obj):
                continue
            for obj2, b2 in enumerate(boxes):
                if (obj1 == obj2):
                    continue
                if (obj2cid2[obj2] == cid1):
                    rid2 = obj2rid[obj2]
                    if (rid1 == rid2):
                        continue
                    for obj3 in rid2obj[rid2]:
                        cid3 = obj2cid2[obj3]
                        if (obj3 in blacklist_obj):
                            continue
                        if (cid1 != cid3):
                            for obj4 in cid2obj2[cid3]:
                                if (obj4 == obj3):
                                    continue
                                if (obj2rid[obj4] == rid1):
                                    min_cid = min(cid1, cid3)
                                    max_cid = max(cid1, cid3)
                                    for obj_iter in cid2obj2[max_cid]:
                                        cid2obj2[min_cid].add(obj_iter)
                                        obj2cid2[obj_iter] = min_cid
                                    cid2obj2[max_cid] = set()
                                    # Features
                                    if_vertical_columns_merged[min_cid] = 1
                                    if_vertical_columns_merged[max_cid] = 0
                                    num_vertical_columns_merged[
                                        min_cid] += num_vertical_columns_merged[
                                            max_cid]
                                    num_vertical_columns_merged[max_cid] = 0
                                    if (if_row_connected[min_cid] == 1
                                            or if_row_connected[max_cid] == 1):
                                        if_row_connected[min_cid] = 1
                                        num_row_connected[
                                            min_cid] += num_row_connected[
                                                max_cid]
                                        num_row_connected[max_cid] = 0
                                        if_row_connected[max_cid] = 0
                                    if (if_connected_by_span[min_cid] == 1 or
                                            if_connected_by_span[max_cid] == 1
                                        ):
                                        if_connected_by_span[min_cid] = 1
                                        num_connected_by_span[
                                            min_cid] += num_connected_by_span[
                                                max_cid]
                                        num_connected_by_span[max_cid] = 0
                                        if_connected_by_span[max_cid] = 0
                                    if (if_connected_by_align[min_cid] == 1 or
                                            if_connected_by_align[max_cid] == 1
                                        ):
                                        if_connected_by_align[min_cid] = 1
                                        num_connected_by_align[
                                            min_cid] += num_connected_by_align[
                                                max_cid]
                                        num_connected_by_align[max_cid] = 0
                                        if_connected_by_align[max_cid] = 0
                                    break
        if (prev_clusters == obj2cid2):
            break
        prev_clusters = obj2cid2

    clusters = [[boxes[i] for i in cluster]
                for cluster in filter(bool, cid2obj2)]
    nodes = [Node(elems) for elems in clusters]
    node_indices = [i for i, x in enumerate(cid2obj2) if x]
    merge_indices = [i for i in range(len(node_indices))]
    page_stat = Node(boxes)
    nodes, merge_indices = merge_nodes(nodes, plane, page_stat, merge_indices)
    # Features
    for idx in range(len(merge_indices)):
        if (merge_indices[idx] != idx):
            cid1 = node_indices[merge_indices[idx]]
            cid2 = node_indices[idx]
            if (if_row_connected[cid1] == 1 or if_row_connected[cid2] == 1):
                if_row_connected[cid1] = 1
                num_row_connected[cid1] += num_row_connected[cid2]
                num_row_connected[cid2] = 0
                if_row_connected[cid2] = 0
            if (if_connected_by_span[cid1] == 1
                    or if_connected_by_span[cid2] == 1):
                if_connected_by_span[cid1] = 1
                num_connected_by_span[cid1] += num_connected_by_span[cid2]
                num_connected_by_span[cid2] = 0
                if_connected_by_span[cid2] = 0
            if (if_connected_by_align[cid1] == 1
                    or if_connected_by_align[cid2] == 1):
                if_connected_by_align[cid1] = 1
                num_connected_by_align[cid1] += num_connected_by_align[cid2]
                num_connected_by_align[cid2] = 0
                if_connected_by_align[cid2] = 0
            if (if_vertical_columns_merged[cid1] == 1
                    or if_vertical_columns_merged[cid2] == 1):
                if_vertical_columns_merged[cid1] = 1
                num_vertical_columns_merged[
                    cid1] += num_vertical_columns_merged[cid2]
                num_vertical_columns_merged[cid2] = 0
                if_vertical_columns_merged[cid2] = 0

    # Get Word Spacing Features
    rid2space = defaultdict(float)
    rid2space_norm = defaultdict(float)
    row_indices = [i for i, x in enumerate(rid2obj) if x]
    for rid in row_indices:
        obj_list = list(rid2obj[rid])
        if (len(obj_list) == 1):
            rid2space[rid] = 0
            continue
        obj_boxes = [boxes[obj].bbox[0] for obj in obj_list]
        sorted_obj_idx = [
            i[0] for i in sorted(enumerate(obj_boxes), key=lambda x: x[1])
        ]
        for obj_idx in range(len(sorted_obj_idx) - 1):
            rid2space[rid] += boxes[obj_list[sorted_obj_idx[obj_idx + 1]]].bbox[2] - \
                              boxes[obj_list[sorted_obj_idx[obj_idx]]].bbox[0]
        rid2space_norm[rid] = rid2space[rid] / (len(obj_list) - 1)

    for idx, node in enumerate(nodes):
        node_idx = node_indices[idx]
        if (merge_indices[idx] == idx):
            obj_list = []
            for idx_iter in range(len(merge_indices)):
                if (merge_indices[idx_iter] == idx):
                    obj_list += list(cid2obj2[node_indices[idx_iter]])
            obj_list = list(set(obj_list))
            rid_list = list(set([obj2rid[obj] for obj in obj_list]))
            for rid in rid_list:
                total_word_space[node_idx] += rid2space[rid]
                avg_word_space_norm[node_idx] += rid2space_norm[rid]
                obj_boxes = [
                    boxes[obj].bbox[0] for obj in rid2obj
                    if obj in cid2obj2[node_idx]
                ]
                sorted_obj_idx = [
                    i[0]
                    for i in sorted(enumerate(obj_boxes), key=lambda x: x[1])
                ]
                for obj_idx in range(len(sorted_obj_idx) - 1):
                    node_space[node_idx] += boxes[obj_list[sorted_obj_idx[obj_idx + 1]]].bbox[2] - \
                                            boxes[obj_list[sorted_obj_idx[obj_idx]]].bbox[0]
                avg_node_space_norm[node_idx] += node_space[node_idx] / (
                    len(obj_boxes) - 1)
            avg_word_space[node_idx] = total_word_space[node_idx] / len(
                rid_list)
            avg_word_space_norm[node_idx] /= len(rid_list)
            avg_node_space[node_idx] = node_space[node_idx] / len(rid_list)
            avg_node_space_norm[node_idx] /= len(rid_list)

    new_nodes = []
    new_node_indices = []
    for idx in range(len(merge_indices)):
        if (merge_indices[idx] == idx):
            new_nodes.append(nodes[idx])
            new_node_indices.append(node_indices[idx])

    nodes = new_nodes
    node_indices = new_node_indices
    # Features
    for idx, node in enumerate(nodes):
        node_idx = node_indices[idx]
        node_bbox = (node.x0, node.y0, node.x1, node.y1)
        for i1, b1 in enumerate(boxes_segments):
            if (intersect(node_bbox, b1.bbox)):
                num_segments[node_idx] += 1
        for i1, b1 in enumerate(boxes_figures):
            if (intersect(node_bbox, b1.bbox)):
                num_figures[node_idx] += 1
        for i1, b1 in enumerate(boxes_curves):
            if (intersect(node_bbox, b1.bbox)):
                num_curves[node_idx] += 1

    tables = []
    table_indices = []
    for idx, node in enumerate(nodes):
        node_idx = node_indices[idx]
        isTable = True
        if node.is_table():
            for elem in node.elems:
                if ("table" in elem.get_text().lower()):
                    continue
                if ((node.width - elem.bbox[2] + elem.bbox[0]) <
                        2 * char_width):
                    isTable = False
            if (isTable):
                tables.append(node)
                table_indices.append(node_idx)

    if (combine == True):
        node_features = [0] * 17
        for idx, node in enumerate(nodes):
            node_idx = node_indices[idx]
            node_features = [
                sum(x) for x in zip(node_features, [
                    if_row_connected[node_idx], num_row_connected[node_idx],
                    if_connected_by_span[node_idx],
                    num_connected_by_span[node_idx],
                    if_connected_by_align[node_idx],
                    num_connected_by_align[node_idx],
                    if_vertical_columns_merged[node_idx],
                    num_vertical_columns_merged[node_idx],
                    num_segments[node_idx], num_curves[node_idx],
                    num_figures[node_idx], total_word_space[node_idx],
                    avg_word_space[node_idx], avg_word_space_norm[node_idx],
                    node_space[node_idx], avg_node_space[node_idx],
                    avg_node_space_norm[node_idx]
                ])
            ]
        return [], node_features
    else:
        table_features = []
        for idx, table in enumerate(tables):
            table_idx = table_indices[idx]
            table_features += [[
                if_row_connected[table_idx], num_row_connected[table_idx],
                if_connected_by_span[table_idx],
                num_connected_by_span[table_idx],
                if_connected_by_align[table_idx],
                num_connected_by_align[table_idx],
                if_vertical_columns_merged[table_idx],
                num_vertical_columns_merged[table_idx],
                num_segments[table_idx], num_curves[table_idx],
                num_figures[table_idx], total_word_space[table_idx],
                avg_word_space[table_idx], avg_word_space_norm[table_idx],
                node_space[table_idx], avg_node_space[table_idx],
                avg_node_space_norm[table_idx]
            ]]
        return tables, table_features


def parse_tree_structure(elems, font_stat, page_num, ref_page_seen, tables,
                         favor_figures):
    boxes_segments = elems.segments
    boxes_curves = elems.curves
    boxes_figures = elems.figures
    page_width = elems.layout.width
    page_height = elems.layout.height
    mentions = elems.mentions

    avg_font_pts = get_most_common_font_pts(elems.mentions, font_stat)
    width = get_page_width(
        mentions + boxes_segments + boxes_figures + boxes_curves)

    try:
        char_width = get_char_width(mentions)
    except:
        char_width = 2

    grid_size = avg_font_pts / 2.0

    # Atomic features and marking initialization
    for i, m in enumerate(mentions + boxes_figures):
        m.id = i
        m.feats = defaultdict(bool)
        prefix = ''
        if isinstance(m, LTTextLine) and m.font_name:
            prefix = m.font_name + '-' + str(m.font_size) + '-'
        # center X coordinate
        m.xc = (m.x0 + m.x1) / 2.0
        m.yc = (m.y0 + m.y1) / 2.0
        # Here we snap the elements to its closest grid line to detect rows/columns
        m.feats[prefix + 'x0'] = m.x0_grid = m.x0 // grid_size
        m.feats[prefix + 'x1'] = m.x1_grid = m.x1 // grid_size
        m.feats[prefix + 'xc'] = m.xc_grid = m.xc // grid_size
        m.feats[prefix + 'yc'] = m.yc_grid = m.yc // grid_size

    #Figures for this page
    figures_page = get_figures(mentions, elems.layout.bbox, page_num,
                               boxes_figures, page_width, page_height)

    #Omit tables that overlap with figures if figures need to be favored
    if favor_figures == "True":
        tables_page = []
        for idx, table in enumerate(tables):
            table_box = tuple(table[3:])
            intersect = False
            for fig in figures_page:
                bool_overlap = (table_box[1] <= fig[6]
                                and fig[4] <= table_box[3]
                                and table_box[0] <= fig[5]
                                and fig[3] <= table_box[2])
                if (bool_overlap):
                    intersect = True
                    break
            if not intersect:
                tables_page.append(table)
    else:
        tables_page = tables

    ##Eliminate tables from these boxes
    boxes = []
    for idx1, box in enumerate(mentions):
        intersect = False
        for idx2, table in enumerate(tables_page):
            table_box = tuple(table[3:])
            bool_overlap = (table_box[1] <= box.bbox[2]
                            and box.bbox[0] <= table_box[3]
                            and table_box[0] <= box.bbox[3]
                            and box.bbox[1] <= table_box[2])
            if (bool_overlap):
                intersect = True
                break
        if (not intersect):
            boxes.append(box)

    text_candidates, ref_page_seen = extract_text_candidates(
        boxes, elems.layout.bbox, avg_font_pts, width, char_width, page_num,
        ref_page_seen, boxes_figures, page_width, page_height)
    text_candidates["figure"] = figures_page
    text_candidates["table"] = tables_page

    #Check overlap with figures if figures are favored
    pruned_text_candidates = {}
    if favor_figures == "True":
        for clust in text_candidates:
            pruned_text_candidates[clust] = []
            for idx, box in enumerate(text_candidates[clust]):
                clust_box = tuple(box[3:])
                intersect = False
                for fig in figures_page:
                    bool_overlap = (clust_box[1] <= fig[6]
                                    and fig[4] <= clust_box[3]
                                    and clust_box[0] <= fig[5]
                                    and fig[3] <= clust_box[2])
                    if (bool_overlap):
                        intersect = True
                        break
                if not intersect:
                    pruned_text_candidates[clust].append(box)
        pruned_text_candidates["figure"] = text_candidates["figure"]
    else:
        pruned_text_candidates = text_candidates

    return pruned_text_candidates, ref_page_seen


def extract_text_candidates(boxes, page_bbox, avg_font_pts, width, char_width,
                            page_num, ref_page_seen, boxes_figures, page_width,
                            page_height):
    #Too many "." in the Table of Content pages - ignore because it takes a lot of time
    if (len(boxes) == 0 or len(boxes) > 3500):
        return {}, False
    plane = Plane(page_bbox)
    plane.extend(boxes)

    #Row level clustering - identify objects that have same horizontal alignment
    rid2obj = [set([i]) for i in range(len(boxes))]  # initialize clusters
    obj2rid = list(range(
        len(boxes)))  # default object map to cluster with its own index
    prev_clusters = obj2rid
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2rid[i1] == obj2rid[i2])):
                    continue
                box1 = b1.bbox
                box2 = b2.bbox
                if ((abs(box1[1] - box2[1]) < 0.11 * avg_font_pts)
                        or ((abs(box1[3] - box2[3]) < 0.11 * avg_font_pts))
                        or (round((box1[1] + box1[3]) / 2) == round(
                            (box2[1] + box2[3]) / 2))):
                    min_i = min(i1, i2)
                    max_i = max(i1, i2)
                    rid1 = obj2rid[min_i]
                    rid2 = obj2rid[max_i]
                    for obj_iter in rid2obj[rid2]:
                        rid2obj[rid1].add(obj_iter)
                        obj2rid[obj_iter] = rid1
                    rid2obj[rid2] = set()
        if (prev_clusters == obj2rid):
            break
        prev_clusters = obj2rid

    cid2obj = [set([i]) for i in range(len(boxes))]  # initialize clusters
    obj2cid = list(range(
        len(boxes)))  # default object map to cluster with its own index
    prev_clusters = obj2cid

    #add the code for merging close text boxes in particular row
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2cid[i1] == obj2cid[i2])):
                    continue
                box1 = b1.bbox
                box2 = b2.bbox
                if (obj2rid[i1] == obj2rid[i2]):
                    if (((b1.bbox[0] < b2.bbox[0]) and
                         ((b2.bbox[0] - b1.bbox[2]) <= 2 * char_width)) or
                        ((b2.bbox[0] < b1.bbox[0]) and
                         ((b1.bbox[0] - b2.bbox[2]) <= 2 * char_width))):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
        if (prev_clusters == obj2cid):
            break
        prev_clusters = obj2cid

    #vertical alignment code
    while (True):
        for i1, b1 in enumerate(boxes):
            for i2, b2 in enumerate(boxes):
                if ((i1 == i2) or (obj2cid[i1] == obj2cid[i2])):
                    continue
                if (b1.bbox[1] < b2.bbox[1]):
                    box1 = b1.bbox
                    box2 = b2.bbox
                elif (b2.bbox[1] < b1.bbox[1]):
                    box1 = b2.bbox
                    box2 = b1.bbox
                else:
                    #horizontally aligned
                    continue
                if (abs((box2[3] - box2[1]) - (box1[3] - box1[1])) >
                        0.5 * avg_font_pts):
                    continue
                if (
                        box2[1] < box1[3]
                        or (box2[1] - box1[1] < 1.5 * avg_font_pts)
                        or (box2[3] - box1[3] < 1.5 * avg_font_pts)
                ):  #can probably do better if we find the average space between words
                    if (abs(box1[0] - box2[0]) < 3 * char_width
                            or abs(box1[2] - box2[2]) < 3 * char_width
                            or (((box1[0] + box1[2]) / 2) == (
                                (box2[0] + box2[2]) / 2))):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        #move all objects from cluster cid2 to cid1
                        #reassign cluster ids for all such objects as well
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
        if (prev_clusters == obj2cid):
            break
        prev_clusters = obj2cid

    #get cluster spans
    cid2span = {}
    for cid in range(len(cid2obj)):
        cid2span[cid] = {}
        cid2span[cid]["min_x"] = float("Inf")
        cid2span[cid]["min_y"] = float("Inf")
        cid2span[cid]["max_x"] = float("-Inf")
        cid2span[cid]["max_y"] = float("-Inf")
        for obj in cid2obj[cid]:
            cid2span[cid]["min_x"] = min(cid2span[cid]["min_x"],
                                         boxes[obj].bbox[0])
            cid2span[cid]["max_x"] = max(cid2span[cid]["max_x"],
                                         boxes[obj].bbox[2])
            cid2span[cid]["min_y"] = min(cid2span[cid]["min_y"],
                                         boxes[obj].bbox[1])
            cid2span[cid]["max_y"] = max(cid2span[cid]["max_y"],
                                         boxes[obj].bbox[3])

    #Don't split up references
    references_bbox = []
    references_cid = set()
    for cid in range(len(cid2obj)):
        if (len(cid2obj[cid]) == 1):
            if (boxes[list(
                    cid2obj[cid])[0]].get_text().lower() == "references"):
                references_bbox = [
                    cid2span[cid]["min_x"], cid2span[cid]["min_y"],
                    cid2span[cid]["max_x"], cid2span[cid]["max_y"]
                ]
                for cid2 in range(len(cid2obj)):
                    if (round(cid2span[cid]["min_x"]) == round(
                            cid2span[cid2]["min_x"]) and
                            cid2span[cid]["max_y"] < cid2span[cid2]["min_y"]):
                        references_cid.add(cid2)
                        cid2span[cid2]["min_x"] = cid2span[cid]["min_x"]
                        cid2span[cid2]["max_x"] = cid2span[cid]["max_x"]

    #get a list of empty cids
    empty_cids = [cid for cid in range(len(cid2obj)) if len(cid2obj[cid]) == 0]
    empty_idx = 0

    #Split paras based on whitespaces - seems to work
    if (ref_page_seen == False):
        for cid in range(len(cid2obj)):
            if (len(cid2obj[cid]) > 0 and cid not in empty_cids
                    and cid not in references_cid):
                cid_maxx = max([boxes[obj].bbox[2] for obj in cid2obj[cid]])
                cid_minx = min([boxes[obj].bbox[0] for obj in cid2obj[cid]])
                rid_list = set([obj2rid[obj] for obj in cid2obj[cid]])
                #Get min_y for each row
                rid_miny = {}
                for rid in rid_list:
                    rid_miny[rid] = min([
                        boxes[obj].bbox[1] if obj in cid2obj[cid] else 10000
                        for obj in rid2obj[rid]
                    ])
                sorted_rid_miny = sorted(
                    list(rid_miny.items()), key=operator.itemgetter(1))
                last_rid = 0
                for i in range(len(sorted_rid_miny) - 1):
                    row1 = sorted_rid_miny[i][0]
                    row2 = sorted_rid_miny[i + 1][0]
                    row1_maxx = max([
                        boxes[obj].bbox[2] if obj in cid2obj[cid] else -1
                        for obj in rid2obj[row1]
                    ])
                    row2_minx = min([
                        boxes[obj].bbox[0] if obj in cid2obj[cid] else 10000
                        for obj in rid2obj[row2]
                    ])
                    if (row1_maxx <= cid_maxx
                            and (row2_minx - char_width) > cid_minx):
                        #split cluster cid
                        new_cid_idx = empty_cids[empty_idx]
                        empty_idx += 1
                        for i_iter in range(last_rid, i + 1):
                            obj_list = [
                                obj
                                for obj in rid2obj[sorted_rid_miny[i_iter][0]]
                                if obj2cid[obj] == cid
                            ]
                            for obj in obj_list:
                                cid2obj[cid].remove(obj)
                                cid2obj[new_cid_idx].add(obj)
                                obj2cid[obj] = new_cid_idx
                        last_rid = i + 1

    clusters = [[boxes[i] for i in cluster]
                for cluster in filter(bool, cid2obj)]
    nodes = [Node(elems) for elems in clusters]
    node_indices = [i for i, x in enumerate(cid2obj) if x]
    merge_indices = [i for i in range(len(node_indices))]
    page_stat = Node(boxes)
    nodes, merge_indices = merge_nodes(nodes, plane, page_stat, merge_indices)

    ##Merging Nodes
    new_nodes = []
    new_node_indices = []
    for idx in range(len(merge_indices)):
        if (merge_indices[idx] == idx):
            new_nodes.append(nodes[idx])
            new_node_indices.append(node_indices[idx])

    #Heuristics for Node type
    ref_nodes = []
    new_ref_page_seen = False
    if (len(references_cid) > 0 or ref_page_seen or references_bbox != []):
        new_ref_page_seen = True
    ref_seen_in_node = False or ref_page_seen
    all_boxes = boxes + boxes_figures
    min_y_page = float("Inf")
    for idx, box in enumerate(all_boxes):
        min_y_page = min(min_y_page, box.bbox[1])
    if page_num == -1:
        #handle title, authors and abstract here
        log.error("TODO: no way to handle title authors abstract yet.")
    else:
        #eliminate header, footer, page number
        #sort other text and classify as header/paragraph
        new_nodes.sort(key=cmp_to_key(xy_reading_order))
        for idx, node in enumerate(new_nodes):
            if (idx < len(new_nodes) - 1):
                if ((round(node.y0) == round(min_y_page)
                     or math.floor(node.y0) == math.floor(min_y_page)) and
                        node.y1 - node.y0 < 2 * avg_font_pts):  #can be header
                    idx_new = idx + 1
                    if idx_new < len(new_nodes) - 1:
                        while (idx_new < len(new_nodes) - 1 and
                               (round(node.y0) == round(new_nodes[idx_new].y0))
                               or (math.floor(node.y0) == math.floor(
                                   new_nodes[idx_new].y0))):
                            idx_new += 1
                    if (idx_new < len(new_nodes) - 1):
                        if (new_nodes[idx_new].y0 - node.y0 >
                                1.5 * avg_font_pts):
                            node.type = "Header"
                            continue
            #get captions - first word is fig/figure/table
            first_elem = None
            for elem in node.elems:
                if (round(elem.bbox[0]) == round(node.x0)
                        and round(elem.bbox[1]) == round(node.y0)):
                    first_elem = elem
                    break
            if (first_elem != None):
                text = first_elem.get_text()
                if (len(text) > 10):
                    text = first_elem.get_text()[0:10]
                if ("Table" in text):
                    node.type = "Table Caption"
                    continue
                if ("Fig" in text or "Figure" in text):
                    node.type = "Figure Caption"
                    continue
                if (first_elem.get_text().lower() == "references"):
                    node.type = "Section Header"
                    ref_seen_in_node = True
                    continue
            if (ref_seen_in_node):
                node.type = "List"
                continue
            if (references_bbox != [] or ref_seen_in_node):
                if (node.y0 > references_bbox[3]
                        and node.x0 <= references_bbox[0]
                        and node.x1 > references_bbox[2]):
                    node.type = "List"
                    continue
            if (node.y1 - node.y0 <= 2.0 * avg_font_pts):  #one lines - section
                node.type = "Section Header"
            else:  #multiple lines - para
                node.type = "Paragraph"

    #handle references
    newer_nodes = []
    ref_indices = [False for idx in range(len(new_nodes))]
    for idx1, node1 in enumerate(new_nodes):
        if (ref_indices[idx1] == True):
            continue
        if (node1.type != "List"):
            newer_nodes.append(node1)
            continue
        x0, y0, x1, y1 = node1.x0, node1.y0, node1.x1, node1.y1
        newer_node = node1
        ref_indices[idx1] = True
        for idx2, node2 in enumerate(new_nodes):
            if (idx1 != idx2):
                if (node2.type == "List" and ref_indices[idx2] == False):
                    if ((node2.x0 <= x0 and node2.x1 >= x0)
                            or (x0 <= node2.x0 and x1 >= node2.x0)):
                        newer_node.merge(node2)
                        ref_indices[idx2] = True
        newer_nodes.append(newer_node)

    #handle figures
    for fig_box in boxes_figures:
        node_fig = Node(fig_box)
        node_fig.type = "Figure"
        newer_nodes.append(node_fig)

    tree = {}
    tree["section_header"] = [(page_num, page_width, page_height) +
                              (node.y0, node.x0, node.y1, node.x1)
                              for node in newer_nodes
                              if node.type == "Section Header"]
    tree["header"] = [(page_num, page_width, page_height) +
                      (node.y0, node.x0, node.y1, node.x1)
                      for node in newer_nodes if node.type == "Header"]
    tree["paragraph"] = [(page_num, page_width, page_height) +
                         (node.y0, node.x0, node.y1, node.x1)
                         for node in newer_nodes if node.type == "Paragraph"]
    # tree["figure"] = [(page_num, page_width, page_height) + (node.y0, node.x0, node.y1, node.x1) for node in newer_nodes if node.type=="Figure"]
    tree["figure_caption"] = [(page_num, page_width, page_height) +
                              (node.y0, node.x0, node.y1, node.x1)
                              for node in newer_nodes
                              if node.type == "Figure Caption"]
    tree["table_caption"] = [(page_num, page_width, page_height) +
                             (node.y0, node.x0, node.y1, node.x1)
                             for node in newer_nodes
                             if node.type == "Table Caption"]
    tree["list"] = [(page_num, page_width, page_height) +
                    (node.y0, node.x0, node.y1, node.x1)
                    for node in newer_nodes if node.type == "List"]
    return tree, new_ref_page_seen


def get_figures(boxes, page_bbox, page_num, boxes_figures, page_width,
                page_height):
    plane = Plane(page_bbox)
    plane.extend(boxes)

    nodes_figures = []

    for fig_box in boxes_figures:
        node_fig = Node(fig_box)
        nodes_figures.append(node_fig)

    merge_indices = [i for i in range(len(nodes_figures))]
    page_stat = Node(boxes)
    nodes, merge_indices = merge_nodes(nodes_figures, plane, page_stat,
                                       merge_indices)

    ##Merging Nodes
    new_nodes = []
    for idx in range(len(merge_indices)):
        if (merge_indices[idx] == idx):
            new_nodes.append(nodes[idx])

    figures = [(page_num, page_width, page_height) +
               (node.y0, node.x0, node.y1, node.x1) for node in new_nodes]
    return figures


def merge_nodes(nodes, plane, page_stat, merge_indices):
    '''
    Merges overlapping nodes
    '''
    # Merge inner boxes to the best outer box
    # nodes.sort(key=Node.area)
    to_be_removed = set()
    for inner_idx in range(len(nodes)):
        inner = nodes[inner_idx]
        outers = []
        outers_indices = []
        for outer_idx in range(len(nodes)):
            outer = nodes[outer_idx]
            if outer is inner or outer in to_be_removed: continue
            if intersect(outer.bbox, inner.bbox):
                outers.append(outer)
                outers_indices.append(outer_idx)
        if not outers: continue
        # Best is defined as min L1 distance to outer center
        best_outer = min(
            outers,
            key=lambda outer: l1(center(outer.bbox), center(inner.bbox)))
        best_outer_idx = outers_indices[outers.index(best_outer)]
        to_be_removed.add(inner)
        best_outer.merge(inner)
        for cid_iter in range(len(merge_indices)):
            if (merge_indices[cid_iter] == merge_indices[inner_idx]):
                merge_indices[cid_iter] = merge_indices[best_outer_idx]
    return nodes, merge_indices


def get_most_common_font_pts(mentions, font_stat):
    '''
    font_stat: Counter object of font sizes
    '''
    # default min font size of 1 pt in case no font present
    most_common_font_size = font_stat.most_common(1)[0][0]
    # Corner case when no text on page
    if not most_common_font_size: return 2.0
    count = 0.01  # avoid division by zero
    height_sum = 0.02  # default to pts 2.0
    for m in mentions:
        if m.font_size == most_common_font_size:
            height_sum += m.height
            count += 1
    return height_sum / count


def get_page_width(boxes):
    xmin = float("Inf")
    xmax = float("-Inf")
    for i, b in enumerate(boxes):
        xmin = min(xmin, b.bbox[0])
        xmax = max(xmax, b.bbox[2])

    return (xmax - xmin)


def get_char_width(boxes):
    box_len_sum = 0
    num_char_sum = 0
    for i, b in enumerate(boxes):
        box_len_sum = box_len_sum + b.bbox[2] - b.bbox[0]
        num_char_sum = num_char_sum + len(b.get_text())
    return box_len_sum / num_char_sum



'''
Created on Oct 21, 2015

@author: xiao


'''
import six  # Python 2-3 compatibility
from collections import namedtuple
import numpy as np

# bbox indices
x0 = 0
y0 = 1
x1 = 2
y1 = 3

class Segment(namedtuple('Segment', ['e','vector'])):
    __slots__ = ()
    @property
    def length(self):
        return self.vector[x0] if self.vector[x0] else self.vector[y0]
    def horizontal(self):
        return bool(self.vector[x0])
    def vertical(self):
        return bool(self.vector[y0])
    def __str__(self, *args, **kwargs):
        return ' '.join(str(x) for x in [self.e, self.vector, self.e.linewidth])

def vectorize(e, tolerance = 0.1):
    '''
    vectorizes the pdf object's bounding box
    min_width is the width under which we consider it a line
    instead of a big rectangle
    '''
    tolerance = max(tolerance,e.linewidth)
    is_high = e.height > tolerance
    is_wide = e.width > tolerance
    # if skewed towards a line
    if is_wide and not is_high:
        return (e.width,0.)
    if is_high and not is_wide:
        return (0.,e.height)

def aligned(e1,e2):
    '''
    alignment is determined by two boxes having one exactly the same
    attribute, which could mean parallel, perpendicularly forming a
    corner etc.
    '''
    return (any(close(c1,c2) for c1,c2 in zip(e1.bbox,e2.bbox)) or
            x_center_aligned(e1, e2) or
            y_center_aligned(e1, e2))

def x_center_aligned(e1,e2):
    return close(e1.x0+e1.x1, e2.x0 + e2.x1)

def x_aligned(a, b):
    return x_center_aligned(a, b) or close(a.x0,b.x0) or close(a.x1, b.x1)

def y_center_aligned(e1,e2):
    return close(e1.y0+e1.y1, e2.y0 + e2.y1)

def close(f1,f2,thres = 2.0):
    return abs(f1-f2) < thres

def origin(bbox):
    return bbox[:2]

def center(bbox):
    return ((bbox[x0] + bbox[x1]) / 2, (bbox[y0] + bbox[y1]) / 2)

def area(bbox):
    return (bbox[x1]-bbox[x0])*(bbox[y1]-bbox[y0])

def l1(c1,c2):
    return sum(abs(v1-v2) for v1, v2 in zip(c1,c2))

def segment_diff(s1,s2):
    '''
    Returns the sum of absolute difference between
    two segments' end points.
    Only perfectly aligned segments return 0
    '''
    return abs(s1[0] - s2[0]) + abs(s1[1] - s2[1])

def bound_bboxes(bboxes):
    '''
    Finds the minimal bbox that contains all given bboxes
    '''
    group_x0 = min(map(lambda l:l[x0],bboxes))
    group_y0 = min(map(lambda l:l[y0],bboxes))
    group_x1 = max(map(lambda l:l[x1],bboxes))
    group_y1 = max(map(lambda l:l[y1],bboxes))
    return (group_x0,group_y0,group_x1,group_y1)

def bound_elems(elems):
    '''
    Finds the minimal bbox that contains all given elems
    '''
    group_x0 = min(map(lambda l:l.x0,elems))
    group_y0 = min(map(lambda l:l.y0,elems))
    group_x1 = max(map(lambda l:l.x1,elems))
    group_y1 = max(map(lambda l:l.y1,elems))
    return (group_x0,group_y0,group_x1,group_y1)

def intersect(a,b):
    '''
    Check if two rectangles intersect
    '''
    if (a[x0] == a[x1] or a[y0] == a[y1]):
        return False
    if (b[x0] == b[x1] or b[y0] == b[y1]):
        return False    
    return a[x0] <= b[x1] and b[x0] <= a[x1] \
        and a[y0] <= b[y1] and b[y0] <= a[y1]

def inside(outer,inner):
    return inner[x0] >= outer[x0] and inner[x1] <= outer[x1] \
        and inner[y0] >= outer[y0] and inner[y0] <= outer[y1]

_stretch_dir = np.array([-1,-1,1,1])
def enlarge(bbox, delta):
    return np.array(bbox) + delta * _stretch_dir

def reading_order(e1,e2):
    '''
    A comparator to sort bboxes from top to bottom, left to right
    '''
    b1 = e1.bbox
    b2 = e2.bbox
    if round(b1[y0]) == round(b2[y0]) or round(b1[y1]) == round(b2[y1]):
        return float_cmp(b1[x0], b2[x0])
    return float_cmp(b1[y0], b2[y0])

def xy_reading_order(e1, e2):
    '''
    A comparator to sort bboxes from left to right, top to bottom
    '''
    b1 = e1.bbox
    b2 = e2.bbox
    if round(b1[x0]) == round(b2[x0]):
        return float_cmp(b1[y0], b2[y0])
    return float_cmp(b1[x0], b2[x0])

def column_order(b1, b2):
    '''
    A comparator that sorts bboxes first by "columns", where a column is made
    up of all bboxes that overlap, then by vertical position in each column.

    b1 = [b1.type, b1.top, b1.left, b1.bottom, b1.right]
    b2 = [b2.type, b2.top, b2.left, b2.bottom, b2.right]
    '''
    (top, left, bottom, right) = (1, 2, 3, 4)
    # TODO(senwu): Reimplement the functionality of this comparator to
    # detect the number of columns, and sort those in reading order.

    # TODO: This is just a simple top to bottom, left to right comparator
    # for now.
    if (round(b1[top]) == round(b2[top]) or
            round(b1[bottom]) == round(b2[bottom])):
        return float_cmp(b1[left], b2[left])
    return float_cmp(b1[top], b2[top])

    #  if((b1[left] >= b2[left] and b1[left] <= b2[right]) or
    #          (b2[left] >= b1[left] and b2[left] <= b1[right])):
    #      return float_cmp(b1[top], b2[top])
    #
    #  # Return leftmost columns first
    #  return float_cmp(b1[left], b2[left])

def float_cmp(f1, f2):
    if f1 > f2:
        return 1
    elif f1 < f2:
        return -1
    else:
        return 0

def merge_intervals(elems, overlap_thres = 2.0):
    '''
    Project in x axis
    Sort by start
    Go through segments and keep max x1

    Return a list of non-overlapping intervals
    '''
    overlap_thres = max(0.0, overlap_thres)
    ordered = sorted(elems, key=lambda e:e.x0)

    intervals = []
    cur = [-overlap_thres,-overlap_thres]
    for e in ordered:
        if e.x0 - cur[1] > overlap_thres:
            # Check interval validity
            if cur[1] > 0.0:
                intervals.append(cur)
            cur = [e.x0,e.x1]
            continue
        cur[1] = max(cur[1],e.x1)
    intervals.append(cur)
    # Freeze the interval to tuples
    return map(tuple,intervals)



'''
Created on Oct 12, 2015
Various routines to work with pdf objects
extracted with PDFminer

@author: xiao
'''
import logging
import os
import re
import six  # Python 2+3 compatibility
import string
from collections import Counter
from pdftotree.img_utils import *
from pdftotree.pdf.vector_utils import *
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTFigure
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser
from pdfminer.utils import apply_matrix_pt
from pdfminer.pdfdocument import PDFDocument
from pdftotree.pdf.layout_utils import *

# Compact wrapper representation for the pdf
PDFElems = namedtuple('PDFElems', ['mentions', 'segments', 'curves', 'figures', 'layout', 'chars'])

log = logging.getLogger(__name__)


class CustomPDFPageAggregator(PDFPageAggregator):
    '''
    A custom version of the default pdf miner stateful draw call
    interpreter. Handles the creation of python object from pdf draw
    calls.
    Changes the way LTCurves are created - break up large polylines
    and rectangles into standard segments.
    '''
    line_only_shape = re.compile('ml+h?')

    def paint_path(self, gstate, stroke, fill, evenodd, path):
        '''
        Converting long paths to small segments each time we m=Move
        or h=ClosePath for polygon
        '''
        shape = ''.join(x[0] for x in path)
        prev_split = 0
        for i in range(len(shape)):
            if shape[i] == 'm' and prev_split != i:
                self.paint_single_path(gstate, stroke, fill, evenodd, path[prev_split:i])
                prev_split = i
            if shape[i] == 'h':
                self.paint_single_path(gstate, stroke, fill, evenodd, path[prev_split:i + 1])
                prev_split = i + 1

        # clean up remaining segments
        if prev_split < len(shape):
            self.paint_single_path(gstate, stroke, fill, evenodd, path[prev_split:])

    def paint_single_path(self, gstate, stroke, fill, evenodd, path):
        '''
        Converting a single path draw command into lines and curves objects
        '''
        if len(path) < 2:
            return
        shape = ''.join(x[0] for x in path)

        pts = []
        for p in path:
            for i in range(1, len(p), 2):
                pts.append(apply_matrix_pt(self.ctm, (p[i], p[i + 1])))

        # Line mode
        if self.line_only_shape.match(shape):
            # check for sloped lines first
            has_slope = False
            for i in range(len(pts) - 1):
                if pts[i][0] != pts[i + 1][0] and pts[i][1] != pts[i + 1][1]:
                    has_slope = True
                    break
            if not has_slope:
                for i in range(len(pts) - 1):
                    self.cur_item.add(LTLine(gstate.linewidth, pts[i], pts[i + 1]))

                # Adding the closing line for a polygon, especially rectangles
                if shape.endswith('h'):
                    self.cur_item.add(LTLine(gstate.linewidth, pts[0], pts[-1]))
                return

        # Add the curve as an arbitrary polyline (belzier curve info is lost here)
        self.cur_item.add(LTCurve(gstate.linewidth, pts))


def analyze_pages(file_name, char_margin=1.0):
    '''
    Input: the file path to the PDF file
    Output: yields the layout object for each page in the PDF
    '''
    # Open a PDF file.
    with open(os.path.realpath(file_name), 'rb') as fp:
        # Create a PDF parser object associated with the file object.
        parser = PDFParser(fp)
        # Create a PDF document object that stores the document structure.
        # Supply the password for initialization.
        document = PDFDocument(parser, password='')
        # Create a PDF resource manager object that stores shared resources.
        rsrcmgr = PDFResourceManager()
        # Set parameters for analysis.
        laparams = LAParams(char_margin=char_margin, word_margin=0.1, detect_vertical=True)
        # Create a PDF page aggregator object.
        device = CustomPDFPageAggregator(rsrcmgr, laparams=laparams)
        # Create a PDF interpreter object.
        interpreter = PDFPageInterpreter(rsrcmgr, device)
        # Process each page contained in the document.
        for page_num, page in enumerate(PDFPage.create_pages(document)):
            try:
                interpreter.process_page(page)
            except OverflowError as oe:
                log.execption(oe, ', skipping page', page_num, 'of', file_name)
                continue
            layout = device.get_result()
            yield layout


def normalize_pdf(layout, scaler):
    '''
    Normalizes pdf object coordinates (bot left) to image
    conventions (top left origin).
    Returns the list of chars and average char size
    '''
    chars = []
    mentions = []
    height = scaler * layout.height
    font_size_counter = Counter()
    # Lines longer than this are included in segments
    pts_thres = 2.0 * scaler
    segments = []
    curves = []
    figures = []

    def processor(m):
        # Normalizes the coordinate system to be consistent with
        # image library conventions (top left as origin)
        if isinstance(m, LTComponent):
            m.set_bbox(normalize_bbox(m.bbox, height, scaler))

            if isinstance(m, LTCurve):
                m.pts = normalize_pts(m.pts, height, scaler)
                # Only keep longer lines here
                if isinstance(m, LTLine) and max(m.width, m.height) > pts_thres:
                    segments.append(m)
                    return
                # Here we exclude straight lines from curves
                curves.append(m)
                return

            if isinstance(m, LTFigure):
                figures.append(m)
                return

            # Collect stats on the chars
            if isinstance(m, LTChar):
                chars.append(m)
                # fonts could be rotated 90/270 degrees
                font_size = _font_size_of(m)
                font_size_counter[font_size] += 1
                return

            if isinstance(m, LTTextLine):
                mention_text = keep_allowed_chars(m.get_text()).strip()
                # Skip empty and invalid lines
                if mention_text:
                    # TODO: add subscript detection and use latex underscore
                    # or superscript
                    m.clean_text = mention_text
                    m.font_name, m.font_size = _font_of_mention(m)
                    mentions.append(m)
                return

        # Also include non character annotations
        if isinstance(m, LTAnno):
            chars.append(m)

    traverse_layout(layout, processor)

    # Resets mention y0 to the first y0 of alphanum character instead of
    # considering exotic unicode symbols and sub/superscripts to reflect
    # accurate alignment info
    for m in mentions:
        # best_y1 = min(c.y1 for c in m if isinstance(c, LTChar))
        alphanum_c = next((c for c in m if c.get_text().isalnum()), None)
        if alphanum_c:
            m.set_bbox((m.x0, alphanum_c.y0, m.x1, alphanum_c.y1))

        #     mentions.sort(key = lambda m: (m.y0,m.x0))
    elems = PDFElems(mentions, segments, curves, figures, layout, chars)
    return elems, font_size_counter


def _print_dict(elem_dict):
    '''
    Print a dict in a readable way
    '''
    for key, value in sorted(elem_dict.iteritems()):
        if isinstance(value, collections.Iterable):
            print(key, len(value))
        else:
            print(key, value)


def _font_size_of(ch):
    if isinstance(ch, LTChar):
        return max(map(abs, ch.matrix[:4]))
    return -1


def _font_of_mention(m):
    '''
    Returns the font type and size of the first alphanumeric
    char in the text or None if there isn't any.
    '''
    for ch in m:
        if isinstance(ch, LTChar) and ch.get_text().isalnum():
            return (ch.fontname, _font_size_of(ch))
    return (None, 0)


# Initialize the set of chars allowed in output
_ascii_allowed = [False] * 128
_forbidden_chars = '\n\t'
for c in string.printable: _ascii_allowed[ord(c)] = True
for c in _forbidden_chars: _ascii_allowed[ord(c)] = False


def _allowed_char(c):
    '''
    Returns whether the given unicode char is allowed in output
    '''
    c = ord(c)
    if c < 0:
        return False
    if c < 128:
        return _ascii_allowed[c]
    # Genereally allow unicodes, TODO: check for unicode control characters
    # characters
    return True


def keep_allowed_chars(text):
    '''
    Cleans the text for output
    '''
    #     print ','.join(str(ord(c)) for c in text)
    return ''.join(' ' if c == '\n' else c for c in text.strip() if _allowed_char(c))



'''
Created on Jun 10, 2016

@author: xiao
'''
import six  # Python 2-3 compatibility
from pdftotree.pdf.vector_utils import bound_elems, bound_bboxes
from collections import Counter, defaultdict
from pdfminer.layout import LTLine, LTTextLine, LTCurve, LTFigure, LTComponent
from pdfminer.utils import Plane
from pdftotree.pdf.layout_utils import is_vline, is_same_row
import numbers
import numpy as np
from pprint import pprint
from pdftotree.pdf import grid
from pdftotree.pdf.grid import Grid

def elem_type(elem):
    if isinstance(elem, LTLine):
        return 'line'
    if isinstance(elem, LTCurve):
        return 'curve'
    if isinstance(elem, LTTextLine):
        return 'text'
    if isinstance(elem, LTFigure):
        return 'figure'
    return 'unkown'

class Node(LTComponent):
    '''
    A rectangular region in the document representing certain local semantics.
    Also holds its data and features.
    '''
    def __init__(self, elems):
        '''
        Constructor
        '''
        self.elems = elems
        self.sum_elem_bbox = 0
        for elem in elems:
            self.sum_elem_bbox = self.sum_elem_bbox + abs((elem.bbox[0]-elem.bbox[2])*(elem.bbox[1]-elem.bbox[3]))
        #     # self.sum_elem_bbox = self.sum_elem_bbox + len(elem.get_text())
        self.table_area_threshold = 0.7
        self.set_bbox(bound_elems(elems))
        # self.table_indicator = True
        self.type_counts = Counter(map(elem_type,elems))
        if(elem_type(elems) not in ["figure", "unknown"]):
            self.feat_counts = Counter(kv for e in elems for kv in six.iteritems(e.feats))
        else:
            self.feat_counts = 0
        self.type = "UNK"

    def merge(self, other):
        self.elems.extend(other.elems)
        self.set_bbox(bound_bboxes([self.bbox, other.bbox]))
        self.type_counts += other.type_counts
        self.feat_counts += other.feat_counts

    def area(self):
        return self.height * self.width

    def is_borderless(self):
        # at least this many segments for a table
        return self.type_counts['line'] < 6

    def is_table(self):
        '''
        Count the node's number of mention al ignment in both axes to determine
        if the node is a table.
        '''
        if self.type_counts['text'] < 6 or 'figure' in self.type_counts: return False
        for e in self.elems:
            # Characters written as curve are usually small, discard diagrams here
            if elem_type(e) == 'curve' and e.height*e.width > 100: return False
        # import re
        # space_re = '\\s+'
        # ws_arr = []
        # whitespace_aligned = False
        # for elem in self.elems:
        #     elem_ws = []
        #     for m in re.finditer(space_re, elem.get_text()):
        #         elem_ws.append(m.start())
        #     # print elem, elem_ws
        #     if(len(elem_ws)>0):
        #         ws_arr.append(elem_ws)
        # # print ws_arr
        # if(len(ws_arr)>0):
        #     count_arr = max([ws_arr.count(i) for i in ws_arr])
        #     if(float(count_arr)/len(ws_arr) > 0.75):
        #         return True
        if((self.sum_elem_bbox/(self.height*self.width)) > self.table_area_threshold):
            return False
        has_many_x_align = False
        has_many_y_align = False
        for k, v in six.iteritems(self.feat_counts):
            font_key = k[0]
            if v >= 2 and '-' in font_key: # Text row or column with more than 2 elements
                if font_key[-2] == 'x': has_many_x_align = True
                if font_key[-2] == 'y': has_many_y_align = True
        return has_many_x_align and has_many_y_align
        # return 0.5

    def get_grid(self):
        '''
        Standardize the layout of the table into grids
        '''
        mentions, lines = _split_text_n_lines(self.elems)
        # Sort mentions in reading order where y values are snapped to half height-sized grid
        mentions.sort(key=lambda m:(m.yc_grid, m.xc))

        grid = Grid(mentions, lines, self)
        return grid

    def _find_vbars_for_row(self, plane, row):
        align_grid_size = sum(m.height for m in row)/2.0/len(row) # half the avg height
        # Find all x_coords of vertical bars crossing this row
        ryc = sum(m.yc for m in row)/len(row) # avg yc
        query_rect = (self.x0, ryc, self.x1, ryc)
        vbars = filter(is_vline, plane.find(query_rect)) # vbars in this row
        vbars = [(v.xc, v.xc_grid) for v in vbars]
        vbars.sort()
        # Group bars less than min cell width apart as one bar
        prev_xc = -1
        clustered_vbars = []
        for xc, xc_grid in vbars:
            if prev_xc < 0 or xc - prev_xc > align_grid_size:
                clustered_vbars.append(xc_grid) # only keep snapped coord
                prev_xc = xc
        return clustered_vbars

    def __str__(self, *args, **kwargs):
        return '\t'.join(r.get_text().encode('utf8','replace') for r in self.elems
                         if isinstance(r, LTTextLine))

#############################################
#    Static utilities
#############################################
def _split_text_n_lines(elems):
    texts = []
    lines = []
    for e in elems:
        if isinstance(e, LTTextLine):
            texts.append(e)
        elif isinstance(e, LTLine):
            lines.append(e)
    return texts, lines


def _left_bar(content, default_val):
    last_bar = default_val
    for _coord, val in content:
        if not isinstance(val,LTTextLine):
            last_bar = val
        yield last_bar

def _right_bar(content, default_val):
    return reversed(list(_left_bar(reversed(content), default_val)))

def _find_col_parent_for_row(content):
    pass

def _get_cols(row_content):
    '''
    Counting the number columns based on the content of this row
    '''
    cols = []
    subcell_col = []
    prev_bar = None
    for _coord, item in row_content:
        if isinstance(item, LTTextLine):
            subcell_col.append(item)
        else:# bar, add column content
            # When there is no content, we count a None column
            if prev_bar:
                bar_ranges = (prev_bar, item)
                col_items = subcell_col if subcell_col else [None]
                cols.extend([bar_ranges,col_items])
            prev_bar = item
            subcell_col = []
    # Remove extra column before first bar
    return cols

def _row_str(row_content):
    def strfy(r):
        if r is None: return 'None'
        if isinstance(r, tuple):
            _c, r = r
        if isinstance(r, LTTextLine):
            return r.get_text().encode('utf8','replace')
        if isinstance(r, numbers.Number):
            return '|'
        return str(r)
    return '\t'.join(strfy(r) for r in row_content)

def _get_rows(mentions):
    curr_row = []
    rows = []
    prev = None
    for m in mentions:
        if not is_same_row(prev, m):
            if curr_row: rows.append(curr_row)
            curr_row = []
        curr_row.append(m)
        prev = m
    # Finish up last row
    if curr_row: rows.append(curr_row)
    return rows

def _one_contains_other(s1, s2):
    '''
    Whether one set contains the other
    '''
    return min(len(s1), len(s2)) == len(s1 & s2)



#!/usr/bin/env python
import pytest
import sys
from tests.context import pdftotree


def test_heuristic_completion():
    """Simply test that parse runs to completion without errors."""
    output = pdftotree.parse("tests/input/paleo.pdf")
    assert output is not None


def test_ml_completion():
    """Simply test that ML-based parse runs without errors."""
    output = pdftotree.parse(
        "tests/input/paleo.pdf", model_path="tests/input/paleo_model.pkl")
    assert output is not None



"""For pip."""
from setuptools import setup
from setuptools import find_packages

try:
    from pypandoc import convert
    long_description = convert('README.md', 'rst')
except (IOError, ImportError):
    print("warning: pypandoc module not found. Could not convert MD to RST.")
    long_description = open('README.md').read()


exec(open('pdftotree/_version.py').read())
setup(
    name='pdftotree',
    version=__version__,
    description='Parse PDFs into HTML-like trees.',
    long_description=long_description,
    packages=find_packages(),
    install_requires=[
        'IPython',
        'future',
        'numpy',
        'pandas',
        'pdfminer.six',
        'six',
        'sklearn',
        'tabula-py',
        'wand',
    ],
    keywords=['pdf', 'parsing', 'html'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    url="https://github.com/HazyResearch/pdftotree",
    scripts=[
        'bin/extract_tree',
        'bin/extract_tables',
    ],
    classifiers=[  # https://pypi.python.org/pypi?:action=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
    ],
    project_urls={
        'Tracker': 'https://github.com/HazyResearch/pdftotree/issues',
        'Source': 'https://github.com/HazyResearch/pdftotree',
    },
    python_requires='>3',
    author="Hazy Research",
    author_email="senwu@cs.stanford.edu",
    license='MIT',
)



from __future__ import division
from builtins import filter
from builtins import str
from builtins import range
from builtins import object
import logging
import numpy as np
from pdfminer.utils import Plane
from pdftotree.ml.features import get_alignment_features, get_lines_features
from pdftotree.pdf.pdf_parsers import parse_layout
from pdftotree.pdf.pdf_utils import normalize_pdf, analyze_pages
from pdftotree.utils.bbox_utils import get_rectangles, compute_iou
from pdftotree.utils.display_utils import pdf_to_img
from pdftotree.utils.lines_utils import extend_horizontal_lines
from pdftotree.utils.lines_utils import extend_vertical_lines
from pdftotree.utils.lines_utils import get_vertical_and_horizontal
from pdftotree.utils.lines_utils import merge_horizontal_lines
from pdftotree.utils.lines_utils import merge_vertical_lines
from pdftotree.utils.lines_utils import reorder_lines
from wand.color import Color
from wand.drawing import Drawing


class TableExtractorML(object):
    """
    Object to extract tables regions from pdf files
    """

    def __init__(self, pdf_file):
        self.log = logging.getLogger(__name__)
        self.pdf_file = pdf_file
        self.elems = {}
        self.font_stats = {}
        self.lines_bboxes = []
        self.alignments_bboxes = []
        self.intersection_bboxes = []
        self.bboxes = []
        self.candidates = []
        self.features = []
        self.iou_thresh = 0.8
        self.scanned = False

    def identify_scanned_page(self, boxes, page_bbox, page_width, page_height):
        plane = Plane(page_bbox)
        plane.extend(boxes)
        cid2obj = [set([i]) for i in range(len(boxes))]  # initialize clusters
        # default object map to cluster with its own index
        obj2cid = list(range(len(boxes)))
        prev_clusters = obj2cid
        while (True):
            for i1, b1 in enumerate(boxes):
                for i2, b2 in enumerate(boxes):
                    box1 = b1.bbox
                    box2 = b2.bbox
                    if (box1[0] == box2[0] and box1[2] == box2[2]
                            and round(box1[3]) == round(box2[1])):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
            if (prev_clusters == obj2cid):
                break
            prev_clusters = obj2cid
        clusters = [[boxes[i] for i in cluster]
                    for cluster in filter(bool, cid2obj)]
        if (len(clusters) == 1 and clusters[0][0].bbox[0] < -0.0
                and clusters[0][0].bbox[1] <= 0
                and abs(clusters[0][0].bbox[2] - page_width) <= 5
                and abs(clusters[0][0].bbox[3] - page_height) <= 5):
            return True
        return False

    def parse(self):
        is_scanned = False
        lin_seg_present = False
        for page_num, layout in enumerate(analyze_pages(self.pdf_file)):
            page_num += 1  # indexes start at 1
            elems, font_stat = normalize_pdf(layout, scaler=1)
            self.elems[page_num] = elems
            self.font_stats[page_num] = font_stat
            # code to detect if the page is scanned
            if (len(elems.segments) > 0):
                lin_seg_present = True
            for fig in elems.figures:
                if (fig.bbox[0] <= 0.0 and fig.bbox[1] <= 0.0
                        and round(fig.bbox[2]) == round(elems.layout.width)
                        and round(fig.bbox[3]) == round(elems.layout.height)):
                    self.log.debug(
                        "{} is scanned because of full-page figure.".format(
                            self.pdf_file))
                    is_scanned = True
            page_scanned = self.identify_scanned_page(
                elems.figures, elems.layout.bbox, elems.layout.width,
                elems.layout.height)
            # doc is scanned if any page is scanned
            if (page_scanned == True):
                self.log.debug(
                    "{} is scanned one of its pages is scanned.".format(
                        self.pdf_file))
                is_scanned = True
        if (is_scanned or not lin_seg_present):
            self.scanned = True

    def get_scanned(self):
        if (len(self.elems) == 0):
            self.parse()
        return self.scanned

    def get_candidates(self):
        if (len(self.elems) == 0):
            self.parse()
        if (self.scanned):
            return []
        for page_num in list(self.elems.keys()):
            page_boxes, page_features = self.get_candidates_and_features_page_num(
                page_num)
            self.candidates += page_boxes
            self.features += list(page_features)
        return self.candidates

    def get_candidates_and_features(self):
        self.parse()
        if (self.scanned):
            self.log.info("{} is scanned.".format(self.pdf_file))
            return [], [], self.scanned
        for page_num in list(self.elems.keys()):
            page_boxes, page_features = self.get_candidates_and_features_page_num(
                page_num)
            self.candidates += page_boxes
            self.features += list(page_features)
        return self.candidates, self.features, self.scanned

    def get_candidates_and_features_page_num(self, page_num):
        elems = self.elems[page_num]
        font_stat = self.font_stats[page_num]
        lines_bboxes = self.get_candidates_lines(page_num, elems)
        boxes = []
        # Filter out bboxes that are zero width or height
        for bbox in lines_bboxes:
            if (bbox[5] - bbox[3] > 0 and bbox[6] - bbox[4] > 0):
                boxes += [bbox]
        alignments_bboxes, alignment_features = self.get_candidates_alignments(
            page_num, elems)
        self.log.info(
            "Page Num: {}, Line bboxes: {}, Alignment bboxes: {}".format(
                page_num, len(lines_bboxes), len(alignments_bboxes)))
        alignment_features += get_alignment_features(lines_bboxes, elems,
                                                     font_stat)
        # Filter out bboxes that are zero width or height
        for bbox in alignments_bboxes:
            if (bbox[5] - bbox[3] > 0 and bbox[6] - bbox[4] > 0):
                boxes += [bbox]
        #  boxes = alignments_bboxes + lines_bboxes
        if len(boxes) == 0:
            return [], []
        lines_features = get_lines_features(boxes, elems)
        features = np.concatenate(
            (np.array(alignment_features), np.array(lines_features)), axis=1)
        return boxes, features

    def get_candidates_lines(self, page_num, elems):
        page_width, page_height = int(elems.layout.width), int(
            elems.layout.height)
        lines = reorder_lines(elems.segments)
        vertical_lines, horizontal_lines = get_vertical_and_horizontal(lines)
        extended_vertical_lines = extend_vertical_lines(horizontal_lines)
        extended_horizontal_lines = extend_horizontal_lines(vertical_lines)
        vertical_lines = merge_vertical_lines(
            sorted(extended_vertical_lines + vertical_lines))
        horizontal_lines = merge_horizontal_lines(
            sorted(extended_horizontal_lines + horizontal_lines))
        rectangles = get_rectangles(
            sorted(vertical_lines), sorted(horizontal_lines))
        return [(page_num, page_width, page_height) + bbox
                for bbox in rectangles]

    def get_candidates_alignments(self, page_num, elems):
        page_width, page_height = int(elems.layout.width), int(
            elems.layout.height)
        font_stat = self.font_stats[page_num]
        try:
            nodes, features = parse_layout(elems, font_stat)
        except:
            nodes, features = [], []
        return [(page_num, page_width, page_height) +
                (node.y0, node.x0, node.y1, node.x1)
                for node in nodes], features

    def get_elems(self):
        return self.elems

    def get_font_stats(self):
        return self.font_stats

    def get_labels(self, gt_tables):
        """
        :param gt_tables: dict, keys are page number and values are list of
                          tables bbox within that page
        :return:
        """
        labels = np.zeros(len(self.candidates))
        for i, candidate in enumerate(self.candidates):
            page_num = candidate[0]
            try:
                tables = gt_tables[page_num]
                for gt_table in tables:
                    page_width, page_height, y0, x0, y1, x1 = gt_table
                    w_ratio = float(candidate[1]) / page_width
                    h_ratio = float(candidate[2]) / page_height
                    rescaled_gt_table = (y0 * h_ratio, x0 * w_ratio,
                                         y1 * h_ratio, x1 * w_ratio)
                    iou = compute_iou(candidate[-4:], rescaled_gt_table)
                    if iou > self.iou_thresh:
                        # candidate region is a table
                        labels[i] = 1
            except KeyError:
                # any of the candidates is a true table, all zero labels
                pass
        return labels

    def display_bounding_boxes(self, page_num, bboxes, alternate_colors=True):
        elems = self.elems[page_num]
        page_width, page_height = int(elems.layout.width), int(
            elems.layout.height)
        img = pdf_to_img(self.pdf_file, page_num, page_width, page_height)
        draw = Drawing()
        draw.fill_color = Color('rgba(0, 0, 0, 0)')
        color = Color('blue')
        draw.stroke_color = color
        for block in bboxes:
            top, left, bottom, right = block[-4:]
            draw.stroke_color = Color('rgba({},{},{}, 1)'.format(
                str(np.random.randint(255)), str(np.random.randint(255)),
                str(np.random.randint(255))))
            draw.rectangle(
                left=float(left),
                top=float(top),
                right=float(right),
                bottom=float(bottom))
        draw(img)
        return img



import json
import logging
import numpy as np
import re
import six  # Python 2-3 compatibility
import tabula
from functools import cmp_to_key
from pdfminer.utils import Plane
from pdftotree.ml.features import get_lines_features, get_mentions_within_bbox
from pdftotree.pdf.layout_utils import *
from pdftotree.pdf.pdf_parsers import parse_layout, parse_tree_structure
from pdftotree.pdf.pdf_utils import normalize_pdf, analyze_pages
from pdftotree.utils.bbox_utils import get_rectangles
from pdftotree.utils.lines_utils import extend_horizontal_lines
from pdftotree.utils.lines_utils import extend_vertical_lines
from pdftotree.utils.lines_utils import get_vertical_and_horizontal
from pdftotree.utils.lines_utils import merge_horizontal_lines
from pdftotree.utils.lines_utils import merge_vertical_lines
from pdftotree.utils.lines_utils import reorder_lines

class TreeExtractor(object):
    """
    Object to extract tree structure from pdf files
    """

    def __init__(self, pdf_file):
        self.log = logging.getLogger(__name__)
        self.pdf_file = pdf_file
        self.elems = {}
        self.font_stats = {}
        self.lines_bboxes = []
        self.alignments_bboxes = []
        self.intersection_bboxes = []
        self.bboxes = []
        self.candidates = []
        self.features = []
        self.iou_thresh = 0.8
        self.scanned = False
        self.tree = {}
        self.html = ""

    def identify_scanned_page(self, boxes, page_bbox, page_width, page_height):
        plane = Plane(page_bbox)
        plane.extend(boxes)
        # initialize clusters
        cid2obj = [set([i]) for i in range(len(boxes))]
        # default object map to cluster with its own index
        obj2cid = list(range(len(boxes)))
        prev_clusters = obj2cid
        while(True):
            for i1, b1 in enumerate(boxes):
                for i2, b2 in enumerate(boxes):
                    box1 = b1.bbox
                    box2 = b2.bbox
                    if(box1[0]==box2[0] and box1[2]==box2[2] and round(box1[3])==round(box2[1])):
                        min_i = min(i1, i2)
                        max_i = max(i1, i2)
                        cid1 = obj2cid[min_i]
                        cid2 = obj2cid[max_i]
                        for obj_iter in cid2obj[cid2]:
                            cid2obj[cid1].add(obj_iter)
                            obj2cid[obj_iter] = cid1
                        cid2obj[cid2] = set()
            if(prev_clusters == obj2cid):
                break
            prev_clusters = obj2cid
        clusters = [[boxes[i] for i in cluster] for cluster in filter(bool,
                    cid2obj)]
        if(len(clusters) == 1 and
           clusters[0][0].bbox[0] < -0.0 and
           clusters[0][0].bbox[1] <= 0 and
           abs(clusters[0][0].bbox[2]-page_width) <= 5 and
           abs(clusters[0][0].bbox[3]-page_height) <= 5):
            return True
        return False

    def parse(self):
        is_scanned = False
        lin_seg_present = False
        for page_num, layout in enumerate(analyze_pages(self.pdf_file)):
            page_num += 1  # indexes start at 1
            elems, font_stat = normalize_pdf(layout, scaler=1)
            self.elems[page_num] = elems
            self.font_stats[page_num] = font_stat
            # code to detect if the page is scanned
            if(len(elems.segments) > 0):
                lin_seg_present = True
            for fig in elems.figures:
                if(fig.bbox[0] <= 0.0 and fig.bbox[1] <= 0.0 and
                   round(fig.bbox[2]) == round(elems.layout.width) and
                   round(fig.bbox[3]) == round(elems.layout.height)):
                    is_scanned = True
            page_scanned = self.identify_scanned_page(elems.figures,
                                                      elems.layout.bbox,
                                                      elems.layout.width,
                                                      elems.layout.height)
            # doc is scanned if any page is scanned
            if(page_scanned):
                is_scanned = True
        if(is_scanned or not lin_seg_present):
            self.scanned = True

    def is_scanned(self):
        if(len(self.elems) == 0):
            self.parse()
        return self.scanned

    def get_tables_page_num(self, page_num):
        page_boxes, _ = self.get_candidates_and_features_page_num(page_num)
        tables = page_boxes
        return tables

    def get_candidates_and_features_page_num(self, page_num):
        elems = self.elems[page_num]
        #  font_stat = self.font_stats[page_num]
        #  lines_bboxes = self.get_candidates_lines(page_num, elems)
        alignments_bboxes, alignment_features = self.get_candidates_alignments(
            page_num, elems)
        # print "Page Num: ", page_num, "Line bboxes: ", len(lines_bboxes), ", Alignment bboxes: ", len(alignments_bboxes)
        # alignment_features += get_alignment_features(lines_bboxes, elems, font_stat)
        boxes = alignments_bboxes  # + lines_bboxes
        if len(boxes) == 0:
            return [], []
        lines_features = get_lines_features(boxes, elems)
        features = np.concatenate((np.array(alignment_features),
                                   np.array(lines_features)), axis=1)
        return boxes, features

    def get_candidates_lines(self, page_num, elems):
        page_width = int(elems.layout.width)
        page_height = int(elems.layout.height)
        lines = reorder_lines(elems.segments)
        vertical_lines, horizontal_lines = get_vertical_and_horizontal(lines)
        extended_vertical_lines = extend_vertical_lines(horizontal_lines)
        extended_horizontal_lines = extend_horizontal_lines(vertical_lines)
        vertical_lines = merge_vertical_lines(sorted(extended_vertical_lines +
                                                     vertical_lines))
        horizontal_lines = merge_horizontal_lines(
            sorted(extended_horizontal_lines + horizontal_lines))
        rects = get_rectangles(sorted(vertical_lines),
                               sorted(horizontal_lines))
        return [(page_num, page_width, page_height) + bbox for bbox in rects]

    def get_candidates_alignments(self, page_num, elems):
        page_width = int(elems.layout.width)
        page_height = int(elems.layout.height)
        font_stat = self.font_stats[page_num]
        try:
            nodes, features = parse_layout(elems, font_stat)
        except Exception as e:
            self.log.execption(e)
            nodes, features = [], []
        return [(page_num, page_width, page_height) + (node.y0, node.x0,
                node.y1, node.x1) for node in nodes], features

    def get_elems(self):
        return self.elems

    def get_font_stats(self):
        return self.font_stats

    def get_tree_structure(self, model, favor_figures):
        tables = {}
        # use heuristics to get tables
        if(model is None):
            for page_num in self.elems.keys():
                tables[page_num] = self.get_tables_page_num(page_num)
        # use ML to get tables
        else:
            for page_num in self.elems.keys():
                table_candidates, candidates_features = \
                    self.get_candidates_and_features_page_num(page_num)
                tables[page_num] = []
                if(len(candidates_features) != 0):
                    table_predictions = model.predict(candidates_features)
                    tables[page_num] = [table_candidates[i] for i in
                                        range(len(table_candidates)) if
                                        table_predictions[i] > 0.5]
        # Manage References - indicator to indicate if reference has been seen
        ref_page_seen = False
        for page_num in self.elems.keys():
            # Get Tree Structure for this page
            self.tree[page_num], ref_page_seen = \
                parse_tree_structure(self.elems[page_num],
                                     self.font_stats[page_num],
                                     page_num,
                                     ref_page_seen,
                                     tables[page_num],
                                     favor_figures)
        return self.tree

    def get_html_tree(self):
        self.html = "<html>"
        for page_num in self.elems.keys():
            page_html = "<div id=" + str(page_num) + ">"
            boxes = []
            for clust in self.tree[page_num]:
                for (pnum, pwidth, pheight, top, left, bottom, right) in self.tree[page_num][clust]:
                    boxes += [[clust.lower().replace(' ', '_'), top, left,
                               bottom, right]]

            # TODO: We need to detect columns and sort acccordingly.
            boxes.sort(key=cmp_to_key(column_order))

            #  from pprint import pprint
            #  pprint(boxes, width=120)
            #  import pdb; pdb.set_trace()

            for box in boxes:
                if(box[0] == "table"):
                    table = box[1:]
                    table_html = self.get_html_table(table, page_num)
                    if six.PY2:
                        page_html += table_html.decode('utf-8')
                    elif six.PY3:
                        page_html += table_html
                elif(box[0] == "figure"):
                    fig_str = [str(i) for i in box[1:]]
                    fig_html = ("<figure bbox=" + ",".join(fig_str) +
                                "></figure>")
                    if six.PY2:
                        page_html += fig_html.decode('utf-8')
                    elif six.PY3:
                        page_html += fig_html
                else:
                    (box_html, char_html, top_html, left_html, bottom_html,
                     right_html) = self.get_html_others(box[1:], page_num)
                    page_html += ("<" + box[0] + " char='" + char_html +
                                  "', top='" + top_html + "', left='" +
                                  left_html + "', bottom='" + bottom_html +
                                  "', right='" + right_html + "'>" + box_html +
                                  "</" + box[0] + ">")
            page_html += "</div>"
            self.html += page_html
        self.html += "</html>"
        return self.html

    def get_word_boundaries(self, mention):
        mention_text = mention.get_text()
        mention_chars = []
        for obj in mention:
            if isinstance(obj, LTChar):
                x0, y0, x1, y1 = obj.bbox
                mention_chars.append([obj.get_text(), y0, x0, y1, x1])
        words = []
        mention_words = mention_text.split()
        char_idx = 0
        for word in mention_words:
            curr_word = [word, float("Inf"), float("Inf"), float("-Inf"),
                         float("-Inf")]
            len_idx = 0
            while len_idx < len(word):
                if mention_chars[char_idx][0] == " ":
                    char_idx += 1
                    continue
                if word[len_idx] != mention_chars[char_idx][0]:
                    self.log.warning("Out of order ({}, {})".format(word, mention_chars[char_idx][0]))
                curr_word[1] = min(curr_word[1], mention_chars[char_idx][1])
                curr_word[2] = min(curr_word[2], mention_chars[char_idx][2])
                curr_word[3] = max(curr_word[3], mention_chars[char_idx][3])
                curr_word[4] = max(curr_word[4], mention_chars[char_idx][4])
                len_idx += len(mention_chars[char_idx][0])
                char_idx += 1
            words.append(curr_word)
        return words

    def get_char_boundaries(self, mention):
        #  mention_text = mention.get_text()
        mention_chars = []
        for obj in mention:
            if isinstance(obj, LTChar):
                x0, y0, x1, y1 = obj.bbox
                mention_chars.append([obj.get_text(), y0, x0, y1, x1])
        return mention_chars

    def get_html_others(self, box, page_num):
        node_html = ""
        top_html = ""
        left_html = ""
        bottom_html = ""
        right_html = ""
        char_html = ""
        sep = " "
        elems = get_mentions_within_bbox(box, self.elems[page_num].mentions)
        elems.sort(key=cmp_to_key(reading_order))
        for elem in elems:
            chars = self.get_char_boundaries(elem)
            for char in chars:
                if six.PY2:
                    temp = char[0].encode('utf-8')
                elif six.PY3:
                    temp = char[0]
                if not re.match(r'[\x00-\x1F]', temp):
                    char_html += char[0].replace("'", '"') + sep
                    top_html += str(char[1]) + sep
                    left_html += str(char[2]) + sep
                    bottom_html += str(char[3]) + sep
                    right_html += str(char[4]) + sep
            words = self.get_word_boundaries(elem)
            for word in words:
                # node_html += "<word top="+str(word[1])+" left="+str(word[2])+" bottom="+str(word[3])+" right="+str(word[4])+">"+str(word[0].encode('utf-8'))+"</word> "
                node_html += word[0] + " "
        return node_html, char_html, top_html, left_html, bottom_html, right_html

    def get_html_table(self, table, page_num):
        table_str = [str(i) for i in table]
        table_json = tabula.read_pdf(self.pdf_file, pages=page_num,
                                     area=table_str, output_format="json")
        table_html = ""
        if (len(table_json) > 0):
            table_html = "<table>"
            for i, row in enumerate(table_json[0]["data"]):
                row_str = "<tr>"
                for j, column in enumerate(row):
                    box = [column["top"], column["left"], column["top"] +
                           column["height"], column["left"] + column["width"]]
                    top_html = ""
                    left_html = ""
                    bottom_html = ""
                    right_html = ""
                    word_html = ""
                    sep = " "
                    elems = get_mentions_within_bbox(box, self.elems[page_num].mentions)
                    elems.sort(key=cmp_to_key(reading_order))
                    word_td = ""
                    for elem in elems:
                        words = self.get_word_boundaries(elem)
                        for word in words:
                            if six.PY2:
                                temp = word[0].encode('utf-8')
                            elif six.PY3:
                                temp = word[0]
                            if not re.match(r'[\x00-\x1F]', temp):
                                word_td += word[0] + " "
                                word_html += str(word[0].encode('utf-8')) + sep
                                top_html += str(word[1]) + sep
                                left_html += str(word[2]) + sep
                                bottom_html += str(word[3]) + sep
                                right_html += str(word[4]) + sep
                    if six.PY2:
                        temp = word_td[:-1].encode("utf-8")
                    elif six.PY3:
                        temp = word_td[:-1]
                    row_str += ("<td word='" + word_html + "', top='" +
                                top_html + "', left='" + left_html +
                                "', bottom='" + bottom_html + "', right='" +
                                right_html + "'>" + temp + "</td>")
                    # row_str += "<td word='"+word_html+"', top='"+top_html+"', left='"+left_html+"', bottom='"+bottom_html+"', right='"+right_html+"'>"+str(column["text"].encode('utf-8'))+"</td>"
                    # row_str += "<td char='"+char_html+"', top="+str(column["top"])+", left="+str(column["left"])+", bottom="+str(column["top"]+column["height"])+", right="+str(column["left"]+column["width"])+">"
                    # row_str += str(column["text"].encode('utf-8'))
                    # row_str += "</td>"
                row_str += "</tr>"
                table_html += row_str
            table_html += "</table>"
        return table_html



#!/usr/bin/env python
from pdftotree._version import __version__

# At the top level, prevent logging output in absense of logging config.
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from pdftotree.core import parse



TOLERANCE = 5


def reorder_lines(lines, tol=TOLERANCE):
    """
    Changes the line coordinates to be given as (top, left, bottom, right)
    :param lines: list of lines coordinates
    :return: reordered list of lines coordinates
    """
    reordered_lines = []
    for line in lines:
        # we divide by tol and multiply by tol to truncate numbers, stairs function
        reordered_lines += [(int(round(line.y0 / tol) * tol), int(round(line.x0 / tol) * tol),
                             int(round(line.y1 / tol) * tol), int(round(line.x1 / tol) * tol))]
    return reordered_lines


def merge_vertical_lines(lines, tol=TOLERANCE):
    """
    This function merges lines segment when they are vertically aligned
    :param lines: list of lines coordinates (top, left, bottom, right)
    :return: list of merged lines coordinates
    """
    if len(lines) == 0:
        return []
    merged_lines = [lines[0]]
    for line in lines[1:]:
        last_line = merged_lines[-1]
        if line[1] == last_line[1]:  # lines are vertically aligned
            if line[0] <= last_line[2] + tol:  # lines intersect
                y0, x0, y1, x1 = merged_lines[-1]
                merged_lines[-1] = (y0, x0, line[2], x1)
            else:
                merged_lines.append(line)  # lines are vertically aligned but do not intersect
        else:
            merged_lines.append(line)
    return merged_lines


def merge_horizontal_lines(lines, tol=TOLERANCE):
    """
    This function merges horizontal lines when they are horizontally aligned
    :param lines: list of lines coordinates (top, left, bottom, right)
    :return: list of merged lines coordinates
    """
    if len(lines) == 0:
        return []
    merged_lines = [lines[0]]
    for line in lines[1:]:
        last_line = merged_lines[-1]
        if line[0] == last_line[0]:  # lines are horizontally aligned
            if line[1] <= last_line[3] + tol:  # lines intersect
                y0, x0, y1, x1 = merged_lines[-1]
                merged_lines[-1] = (y0, x0, y1, line[3])
            else:
                merged_lines.append(line)  # lines are horizontally aligned but do not intersect
        else:
            merged_lines.append(line)
    return merged_lines


def get_vertical_and_horizontal(lines):
    """
    Extracts vertical and horizontal lines lists
    :param lines: list of lines coordinates
    :return: vertical_lines, horitontal_lines (2 lists of coordinates)
    """
    # TODO: add some angle tolerance when lines are not perfectly aligned (eg: scanned pdf)
    vertical_lines = sorted([e for e in lines if e[1] == e[3]], key=lambda tup: (tup[1], tup[0]))
    horitontal_lines = sorted([e for e in lines if e[0] == e[2]])
    if len(vertical_lines) > 0:
        vertical_lines = merge_vertical_lines(vertical_lines)
    if len(horitontal_lines) > 0:
        horitontal_lines = merge_horizontal_lines(horitontal_lines)
    return vertical_lines, horitontal_lines


# def extend_vertical_lines_(vertical_lines, horizontal_lines):
#     j = 0
#     i = 0
#     new_vertical_lines = []
#     while i < len(horizontal_lines) and j < len(vertical_lines):
#         if int(horizontal_lines[i][0]) == int(vertical_lines[j][0]):
#             if int(vertical_lines[j][1]) > int(horizontal_lines[i][1]) and int(vertical_lines[j][1]) < int(
#                     horizontal_lines[i][3]):
#                 v = vertical_lines[j]
#                 h = horizontal_lines[i]
#                 new_vertical_lines.append((h[0], h[1], v[2], h[1]))
#                 new_vertical_lines.append((h[0], h[3], v[2], h[3]))
#                 j += 1
#                 i += 1
#             else:
#                 i += 1
#         elif int(horizontal_lines[i][0]) < int(vertical_lines[j][0]):
#             i += 1
#         else:
#             j += 1
#     return new_vertical_lines


def extend_vertical_lines(horizontal_lines, tol=TOLERANCE):
    widths = {}
    for i, line in enumerate(horizontal_lines):
        try:
            widths[(line[1], line[3])] += [i]
        except KeyError:
            widths[(line[1], line[3])] = [i]
    new_vertical_lines = []
    for (x0, x1) in widths.keys():
        if len(widths[(x0, x1)]) > 1:
            lines = [horizontal_lines[i] for i in widths[(x0, x1)]]
            y0 = min([h[0] for h in lines])
            y1 = max([h[2] for h in lines])
            new_vertical_lines += [(y0, x0, y1, x0), (y0, x1, y1, x1)]
    return new_vertical_lines


def extend_horizontal_lines(vertical_lines, tol=TOLERANCE):
    heights = {}
    for i, line in enumerate(vertical_lines):
        try:
            heights[(line[0], line[2])] += [i]
        except KeyError:
            heights[(line[0], line[2])] = [i]
    new_horizontal_lines = []
    for (y0, y1) in heights.keys():
        if len(heights[(y0, y1)]) > 1:
            lines = [vertical_lines[i] for i in heights[(y0, y1)]]
            x0 = min([h[1] for h in lines])
            x1 = max([h[3] for h in lines])
            new_horizontal_lines += [(y0, x0, y0, x1), (y1, x0, y1, x1)]
    return new_horizontal_lines
